# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  api.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import inspect
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from typing import (
    Any,
    AsyncIterator,
    Callable,
    ContextManager,
    Iterator,
    Optional,
    Sequence,
    Type,
    TypeVar,
    cast,
)

from fastapi import Depends, FastAPI, Request, Response
from fastapi.dependencies.utils import (
    get_parameterless_sub_dependant,
    lenient_issubclass,
)
from fastapi.routing import APIRoute, APIRouter

from .bases import AbstractPage, AbstractParams
from .pagination import Page

T = TypeVar("T")
TAbstractParams = TypeVar("TAbstractParams", covariant=True, bound=AbstractParams)

params_value: ContextVar[AbstractParams] = ContextVar("params_value")
page_type: ContextVar[Type[AbstractPage]] = ContextVar("page_type", default=Page)

response_value: ContextVar[Response] = ContextVar("response_value")
request_value: ContextVar[Request] = ContextVar("request_value")


def resolve_params(params: Optional[AbstractParams] = None) -> AbstractParams:
    if params is None:
        try:
            return params_value.get()
        except LookupError:
            raise RuntimeError("Use params or add_pagination")

    return params


def create_page(items: Sequence[T], total: int, params: AbstractParams) -> AbstractPage[T]:
    return page_type.get().create(items, total, params)


def response() -> Response:
    try:
        return response_value.get()
    except LookupError:
        raise RuntimeError("response context var must be set")


def request() -> Request:
    try:
        return request_value.get()
    except LookupError:
        raise RuntimeError("request context var must be set")


def _ctx_var_with_reset(var: ContextVar, value: Any) -> ContextManager[None]:
    token = var.set(value)

    @contextmanager
    def _reset_ctx() -> Iterator[None]:
        yield

        with suppress(ValueError):
            var.reset(token)

    return _reset_ctx()


def set_page(page: Type[AbstractPage]) -> ContextManager[None]:
    return _ctx_var_with_reset(page_type, page)


def _create_page_dependency(page: Type[AbstractPage]) -> Callable[[], AsyncIterator[None]]:
    async def _set_page_type() -> AsyncIterator[None]:
        with set_page(page):
            yield

    return _set_page_type


def _create_params_dependency(
    params: Type[TAbstractParams],
) -> Callable[[TAbstractParams], AsyncIterator[TAbstractParams]]:
    async def _pagination_params(*args, **kwargs) -> AsyncIterator[params]:  # type: ignore
        val = params(*args, **kwargs)
        with _ctx_var_with_reset(params_value, val):
            yield val

    _pagination_params.__signature__ = inspect.signature(params)  # type: ignore

    return _pagination_params


async def _set_request_response(req: Request, res: Response) -> AsyncIterator[None]:
    with _ctx_var_with_reset(response_value, res), _ctx_var_with_reset(request_value, req):
        yield


async def _marker() -> None:
    pass


ParentT = TypeVar("ParentT", APIRouter, FastAPI)


def _update_route(route: APIRoute) -> None:
    if any(d.call is _marker for d in route.dependant.dependencies):
        return

    if not lenient_issubclass(route.response_model, AbstractPage):
        return

    cls = cast(Type[AbstractPage], route.response_model)

    dependencies = [
        Depends(_marker),
        Depends(_set_request_response),
        Depends(_create_params_dependency(cls.__params_type__)),
        Depends(_create_page_dependency(cls)),
    ]

    route.dependencies.extend(dependencies)
    route.dependant.dependencies.extend(
        get_parameterless_sub_dependant(
            depends=d,
            path=route.path_format,
        )
        for d in dependencies
    )


def add_pagination(parent: ParentT) -> ParentT:
    for route in parent.routes:
        if isinstance(route, APIRoute):
            _update_route(route)

    return parent


__all__ = [
    "add_pagination",
    "create_page",
    "resolve_params",
    "response",
    "request",
    "set_page",
]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  parse.py
@Time    :  2020/9/25 19:55
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import argparse
import json
import warnings
from http.cookies import SimpleCookie
from shlex import split
from urllib.parse import urlparse

from w3lib.http import basic_auth_header

from custard.core.processor import DataKitHelper


class CurlParser(argparse.ArgumentParser, DataKitHelper):
    def error(self, message):
        error_msg = "There was an error parsing the curl command: {}".format(message)
        raise ValueError(error_msg)


# 自定义解析的内容
curl_parser = CurlParser()
curl_parser.add_argument("url")
curl_parser.add_argument("-H", "--header", dest="headers", action="append")
curl_parser.add_argument("-X", "--request", dest="method", default="get")
curl_parser.add_argument("-d", "--data-raw", dest="data", default=None)
curl_parser.add_argument("-f", "--form", "--data-urlencode", dest="form", action="append")
curl_parser.add_argument("-u", "--user", dest="auth")

safe_to_ignore_arguments = [
    ["--location"],
    ["--compressed"],
    # `--compressed` argument is not safe to ignore, but it's included here
    # because the `HttpCompressionMiddleware` is enabled by default
    ["-s", "--silent"],
    ["-v", "--verbose"],
    ["-#", "--progress-bar"],
]
for argument in safe_to_ignore_arguments:
    curl_parser.add_argument(*argument, action="store_true")


def curl_to_request_kwargs(curl_command, ignore_unknown_options=True):
    """Convert a cURL command syntax to Request kwargs.
    :param str curl_command: string containing the curl command
    :param bool ignore_unknown_options: If true, only a warning is emitted when
    cURL options are unknown. Otherwise raises an error. (default: True)
    :return: dictionary of Request kwargs
    """
    curl_args = split(curl_command)
    if curl_args[0] != "curl":
        raise ValueError('A curl command must start with "curl"')
    parsed_args, argv = curl_parser.parse_known_args(curl_args[1:])
    if argv:
        msg = "Unrecognized options: {}".format(", ".join(argv))
        if ignore_unknown_options:
            warnings.warn(msg)
        else:
            raise ValueError(msg)
    url = parsed_args.url
    # curl automatically prepends 'http' if the scheme is missing, but Request
    # needs the scheme to work
    parsed_url = urlparse(url)
    if not parsed_url.scheme:
        url = "http://" + url
    result = {"method": parsed_args.method.upper(), "url": url}
    headers, forms, cookies, content_type = [], [], {}, "application/json"
    for header in parsed_args.headers or ():
        name, val = header.split(":", 1)
        name = name.strip()
        val = val.strip()
        name_title = name.title()
        if name_title == "Content-Type":
            content_type = val
        if name_title == "Cookie":
            for name, morsel in SimpleCookie(val).items():
                cookies[name] = morsel.value
        else:
            headers.append((name, val))
    for form in parsed_args.form or ():
        name, val = form.split("=", 1)
        name = name.strip('"')
        val = val.strip('"')
        forms.append((name, val))
    if parsed_args.auth:
        user, password = parsed_args.auth.split(":", 1)
        headers.append(("Authorization", basic_auth_header(user, password)))
    if headers:
        result["headers"] = headers
    if cookies:
        result["cookies"] = cookies
    if parsed_args.data:
        body_ = DataKitHelper.safely_json_loads(parsed_args.data)
        result["body"] = body_
    if forms:
        if content_type == "application/x-www-form-urlencoded":
            result["data-urlencode"] = forms
        else:
            result["form"] = forms
    # print("---parsed_curl_dumps---", json.dumps(result))
    return result


if __name__ == "__main__":
    get_command = """
                  curl --location --request GET '{{baseUrl}}/api/v1/ps/get?l=22&kj=5' \
                  --header 'ab: 11' \
                  --header 'fg: dfg' \
                  --header "Content-Type: text/plain" \
                  --data-raw '{"s": 0}'
                """
    form_command = """
                curl --location --request POST 'https://xxx/index?act=util&op=inital_checked_staff' \
                --header 'Cookie: PHPSESSID=v93jf5ecmgct5lf2fohe9v6p9s' \
                --form 'member_id="106767"' \
                --form 'key="URJoSEz2Fb8NwzNNGfplbYlolW7wph19UiqdLNkhlq4tdmbV98R9IwB3CyhPSrik"'
                """
    data_urlencode_command = """
                curl --location --request POST 'https://xxx/index?act=util&op=inital_checked_staff' \
                --header 'Cookie: PHPSESSID=v93jf5ecmgct5lf2fohe9v6p9s' \
                --header 'Content-Type: application/x-www-form-urlencoded'
                --data-urlencode 'member_id="106767"' \
                --data-urlencode 'key="URJoSEz2Fb8NwzNNGfplbYlolW7wph19UiqdLNkhlq4tdmbV98R9IwB3CyhPSrik"'
                """
    post_command = """
                curl --location --request POST 'https://xxx?act=util&op=test_staff&api_key=URJoSyhPSrik' \
                --header 'Content-Type: application/json' \
                --header 'Cookie: PHPSESSID=9ec54h4skg93n667bu4gl4jq37; XDEBUG_SESSION=1' \
                --data-raw '{
                    "str": "ABC的235678",
                    "number": 13273298163,
                    "dict": {"dict-1":"dict123","dict-2":"dict567"},
                    "array": [{"list-1":"list123","list-2":"list567"}],
                    "boolean": true,
                    "undefined": null
                }'"""
    get_ = curl_to_request_kwargs(get_command)
    form_ = curl_to_request_kwargs(form_command)
    post_ = curl_to_request_kwargs(post_command)
    data_urlencode_ = curl_to_request_kwargs(data_urlencode_command)



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  area_code.py
@Time    :  2022/5/30 1:35 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  ini配置文件处理
"""

import configparser
import logging

logger = logging.getLogger(__name__)


class IniHandle:
    def __init__(self, filepath=None):
        self.conf = configparser.ConfigParser()
        self.conf.read(filepath, encoding="utf-8")

    def read(self, filepath=None):
        """
        打开指定的ini文件
        :param filepath:
        :return: <configparser.ConfigParser object at 0x0000015940785BA8>
        """
        try:
            conf = configparser.ConfigParser()
            conf.read(filepath, encoding="utf-8")
            return conf
        except Exception as FileNotFoundError:
            logger.error("文件读取失败,请检查%s是否存在,错误信息:%s" % (filepath, FileNotFoundError))

    def has_section(self, section, option=None):
        """
        检查节点
        :param section:
        :return:
        """
        try:
            if option is None:
                self.conf.has_section(section)
            else:
                self.conf.has_option(section, option)
        except Exception as e:
            logger.info("无此节点,错误信息%s" % (e))

    def section(self):
        """
        获取ini文件下所有的section值
        :return:  all_section
        """
        return self.conf.sections()

    def options(self, section):
        """
        获取指定section的所有option的Key
        :return:
        """
        if self.conf.has_section(section):
            return self.conf.options(section)
        else:
            raise ValueError(section)

    def sect_option(self, section):
        """
        获取指定section下的option的键值对
        :return: List形式的 [('a', 'b'),('aa', 'bb')]
        """
        if self.conf.has_section(section):
            return self.conf.items(section)
        return None

    def get(self, node, key):
        """
        获取指定section下option的value值
        :param filepath 需要读取的文件
        :param node 父类节点
        :param key  所需要查询内容的单一key
        :return: result 返回对应key的value值
        """
        return self.conf.get(node, key)

    def opt_all(self):
        """
        打印配置文件所有的值(该方法并不是很常用)
        :return:
        """
        for section in self.section():
            logger.info("[" + section + "]")
            for K, V in self.conf.items(section):
                logger.info(K + "=" + V)

    def sections_to_dict(self):
        """
        读取所有section到字典中
        :return:
        """
        res_1, res_2 = {}, {}
        sections = self.conf.sections()
        for sec in sections:
            for key, val in self.conf.items(sec):
                res_2[key] = val
            res_1[sec] = res_2.copy()
            res_2.clear()
        return res_1

    def remove_option(self, section, key=None):
        """
        删除一个 section中的一个item(以键值KEY为标识)
        :param section:
        :param key:
        :return:
        """
        if key is None:
            self.has_section(section)
            self.conf.remove_section(section)
        else:
            self.has_section(section, key)
            self.conf.remove_option(section, key)

    def add_section(self, section):
        """
        添加一个section
        :param section:
        :return:
        """
        self.conf.add_section(section)

    def set(self, section, key, value):
        """
        往section添加key和value
        :param section:
        :param key:
        :param value:
        :return:
        """
        self.conf.set(section, key, value)



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  exceptions.py
@Time    :  2022/5/15 10:37 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""

__all__ = ("FunctionTimedOut", "RETRY_SAME_TIMEOUT")

RETRY_SAME_TIMEOUT = "RETRY_SAME_TIMEOUT"


class TimeOutError(Exception):
    """
    An operation timed out
    """


class FunctionTimedOut(Exception):
    def __init__(
        self,
        msg="",
        timed_out_after=None,
        timed_out_function=None,
        timed_out_args=None,
        timed_out_kwargs=None,
    ):
        """
        Args:
            msg:
            timed_out_after:
            timed_out_function:
            timed_out_args:
            timed_out_kwargs:
        """
        self.timed_out_after = timed_out_after
        self.timed_out_function = timed_out_function
        self.timed_out_args = timed_out_args
        self.timed_out_kwargs = timed_out_kwargs

        if not msg:
            msg = self.get_msg()

        BaseException.__init__(self, msg)

        self.msg = msg

    def get_msg(self):
        if self.timed_out_function is not None:
            timed_out_func_name = self.timed_out_function.__name__
        else:
            timed_out_func_name = "Unknown Function"
        timed_out_after_str = "%f" % (self.timed_out_after,) if self.timed_out_after is not None else "Unknown"

        return "Function %s (args=%s) (kwargs=%s) timed out after %s seconds.\n" % (
            timed_out_func_name,
            repr(self.timed_out_args),
            repr(self.timed_out_kwargs),
            timed_out_after_str,
        )



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  useragent.py
@Time    :  2022/5/27 1:50 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import bisect
import random as random_module
import re
import string
from calendar import timegm
from collections import OrderedDict
from datetime import date, datetime, timedelta
from datetime import date as dtdate
from typing import Collection, Dict, Iterable, Sequence, TypeVar, Union

from dateutil.tz import tzutc

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

T = TypeVar("T")
ElementsType = Collection[T]
DateParseType = Union[date, datetime, timedelta, str, int]
HueType = TypeVar("HueType", str, float, Sequence[int])
GenderType = TypeVar("GenderType", bound=Literal["M", "F"])

_re_hash = re.compile(r"#")
_re_perc = re.compile(r"%")
_re_excl = re.compile(r"!")
_re_at = re.compile(r"@")
_re_qm = re.compile(r"\?")
_re_cir = re.compile(r"\^")

__use_weighting__ = False

userAgents: ElementsType = (
    "chrome",
    "firefox",
    "internet_explorer",
    "opera",
    "safari",
)

win_platform_tokens: ElementsType = (
    "win 95",
    "win 98",
    "win 98; Win 9x 5.90",
    "win CE",
    "win NT 5.0",
    "win NT 5.01",
    "win NT 5.1",
    "win NT 5.2",
    "win NT 6.0",
    "win NT 6.1",
    "win NT 6.2",
    "win NT 10.0",
)

linux_processors: ElementsType = ("i686", "x86_64")

mac_processors: ElementsType = ("Intel", "PPC", "U; Intel", "U; PPC")

android_versions: ElementsType = (
    "5.0",
    "5.0.1",
    "5.0.2",
    "5.1",
    "5.1.1",
    "6.0",
    "6.0.1",
    "7.0",
    "7.1",
    "7.1.1",
    "7.1.2",
    "8.0.0",
    "8.1.0",
    "9",
    "10",
    "11",
)

apple_devices: ElementsType = ("iPhone", "iPad")

ios_versions: ElementsType = (
    "3.1.3",
    "5.1.1",
    "6.1.6",
    "7.1.2",
    "9.3.5",
    "9.3.6",
    "10.3.3",
    "10.3.5",
    "12.5.8",
    "15.2",
    "15.2.1",
)

random = random_module.Random()
mod_random = random  # compat with name released in 0.8


def random_sample(random=None) -> float:
    if random is None:
        random = mod_random
    return random.uniform(0.0, 1.0)


def cumsum(it: Iterable[float]):
    total: float = 0
    for x in it:
        total += x
        yield total


def choices_distribution_unique(a, p, random=None, length: int = 1):
    # As of Python 3.7, there isn't a way to sample unique elements that takes
    # weight into account.
    if random is None:
        random = mod_random

    assert p is not None
    assert len(a) == len(p)
    assert len(a) >= length, "You can't request more unique samples than elements in the dataset."

    choices = []
    items = list(a)
    probabilities = list(p)
    for i in range(length):
        cdf = tuple(cumsum(probabilities))
        normal = cdf[-1]
        cdf2 = [i / normal for i in cdf]
        uniform_sample = random_sample(random=random)
        idx = bisect.bisect_right(cdf2, uniform_sample)
        item = items[idx]
        choices.append(item)
        probabilities.pop(idx)
        items.pop(idx)
    return choices


def choices_distribution(a, p, random=None, length: int = 1):
    if random is None:
        random = mod_random

    if p is not None:
        assert len(a) == len(p)

    if hasattr(random, "choices"):
        if length == 1 and p is None:
            return [random.choice(a)]
        else:
            return random.choices(a, weights=p, k=length)
    else:
        choices = []

        if p is None:
            p = itertools.repeat(1, len(a))  # type: ignore

        cdf = list(cumsum(p))  # type: ignore
        normal = cdf[-1]
        cdf2 = [i / normal for i in cdf]
        for _i in range(length):
            uniform_sample = random_sample(random=random)
            idx = bisect.bisect_right(cdf2, uniform_sample)
            item = a[idx]
            choices.append(item)
        return choices


def random_elements(
    elements: ElementsType = ("a", "b", "c"),
    length=None,
    unique: bool = False,
    use_weighting=None,
):
    use_weighting = use_weighting if use_weighting is not None else __use_weighting__

    if isinstance(elements, dict) and not isinstance(elements, OrderedDict):
        raise ValueError("Use OrderedDict only to avoid dependency on PYTHONHASHSEED (See #363).")

    fn = choices_distribution_unique if unique else choices_distribution

    if length is None:
        length = random.randint(1, len(elements))

    if unique and length > len(elements):
        raise ValueError("Sample length cannot be longer than the number of unique elements to pick from.")

    if isinstance(elements, dict):
        if not hasattr(elements, "_key_cache"):
            elements._key_cache = tuple(elements.keys())  # type: ignore

        choices = elements._key_cache  # type: ignore[attr-defined]
        probabilities = tuple(elements.values()) if use_weighting else None
    else:
        if unique:
            # shortcut
            return random.sample(elements, length)
        choices = elements
        probabilities = None

    return fn(
        tuple(choices),
        probabilities,
        random,
        length=length,
    )


def mac_processor() -> str:
    """Generate a MacOS processor token used in user agent strings."""
    return random_element(mac_processors)


def linux_processor() -> str:
    """Generate a Linux processor token used in user agent strings."""
    return random_element(linux_processors)


user_agents: ElementsType = (
    "chrome",
    "firefox",
    "internet_explorer",
    "opera",
    "safari",
)


def user_agent() -> str:
    name: str = random_element(user_agents)
    print(name)


def chrome(
    version_from: int = 13,
    version_to: int = 63,
    build_from: int = 800,
    build_to: int = 899,
) -> str:
    """Generate a Chrome web browser user agent string."""
    saf: str = f"{random.randint(531, 536)}.{random.randint(0, 2)}"
    bld: str = lexify(numerify("##?###"), string.ascii_uppercase)
    tmplt: str = "({0}) AppleWebKit/{1} (KHTML, like Gecko)" " Chrome/{2}.0.{3}.0 Safari/{4}"
    tmplt_ios: str = "({0}) AppleWebKit/{1} (KHTML, like Gecko)" " CriOS/{2}.0.{3}.0 Mobile/{4} Safari/{1}"
    platforms: ElementsType = (
        tmplt.format(
            linux_platform_token(),
            saf,
            random.randint(version_from, version_to),
            random.randint(build_from, build_to),
            saf,
        ),
        tmplt.format(
            win_platform_token(),
            saf,
            random.randint(version_from, version_to),
            random.randint(build_from, build_to),
            saf,
        ),
        tmplt.format(
            mac_platform_token(),
            saf,
            random.randint(version_from, version_to),
            random.randint(build_from, build_to),
            saf,
        ),
        tmplt.format(
            "Linux; {}".format(android_platform_token()),
            saf,
            random.randint(version_from, version_to),
            random.randint(build_from, build_to),
            saf,
        ),
        tmplt_ios.format(
            ios_platform_token(),
            saf,
            random.randint(version_from, version_to),
            random.randint(build_from, build_to),
            bld,
        ),
    )
    return "Mozilla/5.0 " + random_element(platforms)


language_locale_codes = {
    "aa": ("DJ", "ER", "ET"),
    "af": ("ZA",),
    "ak": ("GH",),
    "am": ("ET",),
    "an": ("ES",),
    "apn": ("IN",),
    "ar": (
        "AE",
        "BH",
        "DJ",
        "DZ",
        "EG",
        "EH",
        "ER",
        "IL",
        "IN",
        "IQ",
        "JO",
        "KM",
        "KW",
        "LB",
        "LY",
        "MA",
        "MR",
        "OM",
        "PS",
        "QA",
        "SA",
        "SD",
        "SO",
        "SS",
        "SY",
        "TD",
        "TN",
        "YE",
    ),
    "as": ("IN",),
    "ast": ("ES",),
    "ayc": ("PE",),
    "az": ("AZ", "IN"),
    "be": ("BY",),
    "bem": ("ZM",),
    "ber": ("DZ", "MA"),
    "bg": ("BG",),
    "bhb": ("IN",),
    "bho": ("IN",),
    "bn": ("BD", "IN"),
    "bo": ("CN", "IN"),
    "br": ("FR",),
    "brx": ("IN",),
    "bs": ("BA",),
    "byn": ("ER",),
    "ca": ("AD", "ES", "FR", "IT"),
    "ce": ("RU",),
    "ckb": ("IQ",),
    "cmn": ("TW",),
    "crh": ("UA",),
    "cs": ("CZ",),
    "csb": ("PL",),
    "cv": ("RU",),
    "cy": ("GB",),
    "da": ("DK",),
    "de": ("AT", "BE", "CH", "DE", "LI", "LU"),
    "doi": ("IN",),
    "dv": ("MV",),
    "dz": ("BT",),
    "el": ("GR", "CY"),
    "en": (
        "AG",
        "AU",
        "BW",
        "CA",
        "DK",
        "GB",
        "HK",
        "IE",
        "IN",
        "NG",
        "NZ",
        "PH",
        "SG",
        "US",
        "ZA",
        "ZM",
        "ZW",
    ),
    "eo": ("US",),
    "es": (
        "AR",
        "BO",
        "CL",
        "CO",
        "CR",
        "CU",
        "DO",
        "EC",
        "ES",
        "GT",
        "HN",
        "MX",
        "NI",
        "PA",
        "PE",
        "PR",
        "PY",
        "SV",
        "US",
        "UY",
        "VE",
    ),
    "et": ("EE",),
    "eu": ("ES", "FR"),
    "fa": ("IR",),
    "ff": ("SN",),
    "fi": ("FI",),
    "fil": ("PH",),
    "fo": ("FO",),
    "fr": ("CA", "CH", "FR", "LU"),
    "fur": ("IT",),
    "fy": ("NL", "DE"),
    "ga": ("IE",),
    "gd": ("GB",),
    "gez": ("ER", "ET"),
    "gl": ("ES",),
    "gu": ("IN",),
    "gv": ("GB",),
    "ha": ("NG",),
    "hak": ("TW",),
    "he": ("IL",),
    "hi": ("IN",),
    "hne": ("IN",),
    "hr": ("HR",),
    "hsb": ("DE",),
    "ht": ("HT",),
    "hu": ("HU",),
    "hy": ("AM",),
    "ia": ("FR",),
    "id": ("ID",),
    "ig": ("NG",),
    "ik": ("CA",),
    "is": ("IS",),
    "it": ("CH", "IT"),
    "iu": ("CA",),
    "iw": ("IL",),
    "ja": ("JP",),
    "ka": ("GE",),
    "kk": ("KZ",),
    "kl": ("GL",),
    "km": ("KH",),
    "kn": ("IN",),
    "ko": ("KR",),
    "kok": ("IN",),
    "ks": ("IN",),
    "ku": ("TR",),
    "kw": ("GB",),
    "ky": ("KG",),
    "lb": ("LU",),
    "lg": ("UG",),
    "li": ("BE", "NL"),
    "lij": ("IT",),
    "ln": ("CD",),
    "lo": ("LA",),
    "lt": ("LT",),
    "lv": ("LV",),
    "lzh": ("TW",),
    "mag": ("IN",),
    "mai": ("IN",),
    "mg": ("MG",),
    "mhr": ("RU",),
    "mi": ("NZ",),
    "mk": ("MK",),
    "ml": ("IN",),
    "mn": ("MN",),
    "mni": ("IN",),
    "mr": ("IN",),
    "ms": ("MY",),
    "mt": ("MT",),
    "my": ("MM",),
    "nan": ("TW",),
    "nb": ("NO",),
    "nds": ("DE", "NL"),
    "ne": ("NP",),
    "nhn": ("MX",),
    "niu": ("NU", "NZ"),
    "nl": ("AW", "BE", "NL"),
    "nn": ("NO",),
    "nr": ("ZA",),
    "nso": ("ZA",),
    "oc": ("FR",),
    "om": ("ET", "KE"),
    "or": ("IN",),
    "os": ("RU",),
    "pa": ("IN", "PK"),
    "pap": ("AN", "AW", "CW"),
    "pl": ("PL",),
    "ps": ("AF",),
    "pt": ("BR", "PT"),
    "quz": ("PE",),
    "raj": ("IN",),
    "ro": ("RO",),
    "ru": ("RU", "UA"),
    "rw": ("RW",),
    "sa": ("IN",),
    "sat": ("IN",),
    "sc": ("IT",),
    "sd": ("IN", "PK"),
    "se": ("NO",),
    "shs": ("CA",),
    "si": ("LK",),
    "sid": ("ET",),
    "sk": ("SK",),
    "sl": ("SI",),
    "so": ("DJ", "ET", "KE", "SO"),
    "sq": ("AL", "ML"),
    "sr": ("ME", "RS"),
    "ss": ("ZA",),
    "st": ("ZA",),
    "sv": ("FI", "SE"),
    "sw": ("KE", "TZ"),
    "szl": ("PL",),
    "ta": ("IN", "LK"),
    "tcy": ("IN",),
    "te": ("IN",),
    "tg": ("TJ",),
    "th": ("TH",),
    "the": ("NP",),
    "ti": ("ER", "ET"),
    "tig": ("ER",),
    "tk": ("TM",),
    "tl": ("PH",),
    "tn": ("ZA",),
    "tr": ("CY", "TR"),
    "ts": ("ZA",),
    "tt": ("RU",),
    "ug": ("CN",),
    "uk": ("UA",),
    "unm": ("US",),
    "ur": ("IN", "PK"),
    "uz": ("UZ",),
    "ve": ("ZA",),
    "vi": ("VN",),
    "wa": ("BE",),
    "wae": ("CH",),
    "wal": ("ET",),
    "wo": ("SN",),
    "xh": ("ZA",),
    "yi": ("US",),
    "yo": ("NG",),
    "yue": ("HK",),
    "zh": ("CN", "HK", "SG", "TW"),
    "zu": ("ZA",),
}


def languageCode():
    return random_element(language_locale_codes.keys())


def localeLang():
    language_code = languageCode()
    return (
        language_code
        + "_"
        + random_element(
            language_locale_codes[language_code],
        )
    )


def random_int(min: int = 0, max: int = 9999, step: int = 1) -> int:
    return random.randrange(min, max + 1, step)


def random_digit() -> int:
    """Generate a random digit (0 to 9)."""

    return random.randint(0, 9)


def random_digit_not_null() -> int:
    return random.randint(1, 9)


def random_digit_or_empty() -> Union[int, str]:
    if random.randint(0, 1):
        return random.randint(0, 9)
    return ""


def random_digit_not_null_or_empty() -> Union[int, str]:
    if random.randint(0, 1):
        return random.randint(1, 9)
    else:
        return ""


def random_number(digits=None, fix_len: bool = False) -> int:
    if digits is None:
        digits = random_digit_not_null()
    if digits < 0:
        raise ValueError("The digit parameter must be greater than or equal to 0.")
    if fix_len:
        if digits > 0:
            return random.randint(pow(10, digits - 1), pow(10, digits) - 1)
        raise ValueError("A number of fixed length cannot have less than 1 digit in it.")
    return random.randint(0, pow(10, digits) - 1)


def random_letter() -> str:
    return random.choice(getattr(string, "letters", string.ascii_letters))


def random_choices(self, elements: ElementsType = ("a", "b", "c"), length: int = None) -> Sequence[T]:
    return self.random_elements(elements, length, unique=False)


def random_letters(length: int = 16) -> Sequence[str]:
    return random_choices(
        getattr(string, "letters", string.ascii_letters),
        length=length,
    )


def rand_lowercase_letter() -> str:
    return random.choice(string.ascii_lowercase)


def rand_uppercase_letter() -> str:
    return random.choice(string.ascii_uppercase)


def numerify(text: str = "###") -> str:
    text = _re_hash.sub(lambda x: str(random_digit()), text)
    text = _re_perc.sub(lambda x: str(random_digit_not_null()), text)
    text = _re_excl.sub(lambda x: str(random_digit_or_empty()), text)
    text = _re_at.sub(lambda x: str(random_digit_not_null_or_empty()), text)
    return text


def lexify(text: str = "????", letters: str = string.ascii_letters) -> str:
    return _re_qm.sub(lambda x: random_element(letters), text)


def datetime_to_timestamp(dt: Union[dtdate, datetime]) -> int:
    if isinstance(dt, datetime) and getattr(dt, "tzinfo", None) is not None:
        dt = dt.astimezone(tzutc())
    return timegm(dt.timetuple())


def parse_date_string(cls, value: str) -> Dict[str, float]:
    parts = cls.regex.match(value)
    if not parts:
        raise Exception(f"Can't parse date string `{value}`")
    parts = parts.groupdict()
    time_params: Dict[str, float] = {}
    for name_, param_ in parts.items():
        if param_:
            time_params[name_] = int(param_)

    if "years" in time_params:
        if "days" not in time_params:
            time_params["days"] = 0
        time_params["days"] += 365.24 * time_params.pop("years")
    if "months" in time_params:
        if "days" not in time_params:
            time_params["days"] = 0
        time_params["days"] += 30.42 * time_params.pop("months")

    if not time_params:
        raise Exception(f"Can't parse date string `{value}`")
    return time_params


def _parse_date_time(value: DateParseType, tzinfo=None) -> int:
    if isinstance(value, (datetime, dtdate)):
        return datetime_to_timestamp(value)
    now = datetime.now(tzinfo)
    if isinstance(value, timedelta):
        return datetime_to_timestamp(now + value)
    if isinstance(value, str):
        if value == "now":
            return datetime_to_timestamp(datetime.now(tzinfo))
        time_params = parse_date_string(value)
        return datetime_to_timestamp(now + timedelta(**time_params))  # type: ignore
    if isinstance(value, int):
        return datetime_to_timestamp(now + timedelta(value))
    raise Exception(f"Invalid format for date {value!r}")


def _parse_date(cls, value: DateParseType) -> dtdate:
    if isinstance(value, datetime):
        return value.date()
    elif isinstance(value, dtdate):
        return value
    today = dtdate.today()
    if isinstance(value, timedelta):
        return today + value
    if isinstance(value, str):
        if value in ("today", "now"):
            return today
        time_params = cls._parse_date_string(value)
        return today + timedelta(**time_params)  # type: ignore
    if isinstance(value, int):
        return today + timedelta(value)
    raise Exception(f"Invalid format for date {value!r}")


def date_time_between(
    start_date: DateParseType = "-30y",
    end_date: DateParseType = "now",
    tzinfo=None,
) -> datetime:
    start_date = _parse_date_time(start_date, tzinfo=tzinfo)
    end_date = _parse_date_time(end_date, tzinfo=tzinfo)
    ts = start_date + random.random() if end_date - start_date <= 1 else random.randint(start_date, end_date)
    if tzinfo is None:
        return datetime(1970, 1, 1, tzinfo=tzinfo) + timedelta(seconds=ts)
    else:
        return (datetime(1970, 1, 1, tzinfo=tzutc()) + timedelta(seconds=ts)).astimezone(tzinfo)


def firefox() -> str:
    """Generate a Mozilla Firefox web browser user agent string."""
    ver: ElementsType = (
        (f"Gecko/{date_time_between(datetime(2011, 1, 1))} " f"Firefox/{random.randint(4, 15)}.0"),
        (f"Gecko/{date_time_between(datetime(2010, 1, 1))} " f"Firefox/3.6.{random.randint(1, 20)}"),
        f"Gecko/{date_time_between(datetime(2010, 1, 1))} Firefox/3.8",
    )
    tmplt_win: str = "({0}; {1}; rv:1.9.{2}.20) {3}"
    tmplt_lin: str = "({0}; rv:1.9.{1}.20) {2}"
    tmplt_mac: str = "({0}; rv:1.9.{1}.20) {2}"
    tmplt_and: str = "({0}; Mobile; rv:{1}.0) Gecko/{1}.0 Firefox/{1}.0"
    tmplt_ios: str = "({0}) AppleWebKit/{1} (KHTML, like Gecko) FxiOS/{2}.{3}.0 Mobile/{4} Safari/{1}"
    saf: str = "{}.{}".format(random.randint(531, 536), random.randint(0, 2))
    bld: str = lexify(numerify("##?###"), string.ascii_uppercase)
    bld2: str = lexify(numerify("#?####"), string.ascii_lowercase)
    platforms: ElementsType = (
        tmplt_win.format(
            win_platform_token(),
            localeLang().replace("_", "-"),
            random.randint(0, 2),
            random.choice(ver),
        ),
        tmplt_lin.format(
            linux_platform_token(),
            random.randint(5, 7),
            random.choice(ver),
        ),
        tmplt_mac.format(
            mac_platform_token(),
            random.randint(2, 6),
            random.choice(ver),
        ),
        tmplt_and.format(android_platform_token(), random.randint(5, 68)),
        tmplt_ios.format(
            ios_platform_token(),
            saf,
            random.randint(9, 18),
            bld2,
            bld,
        ),
    )

    return "Mozilla/5.0 " + random_element(platforms)


def localeLang() -> str:
    language_code = languageCode()
    return (
        language_code
        + "_"
        + random_element(
            language_locale_codes[language_code],
        )
    )


def safari() -> str:
    saf: str = f"{random.randint(531, 535)}." f"{random.randint(1, 50)}." f"{random.randint(1, 7)}"

    ver: str = (
        f"{random.randint(4, 5)}.{random.randint(0, 1)}"
        if not random.getrandbits(1)
        else f"{random.randint(4, 5)}.0.{random.randint(1, 5)}"
    )

    tmplt_win: str = "(win; U; {0}) AppleWebKit/{1} (KHTML, like Gecko)" " Version/{2} Safari/{3}"
    tmplt_mac: str = "({0} rv:{1}.0; {2}) AppleWebKit/{3} (KHTML, like Gecko)" " Version/{4} Safari/{5}"
    tmplt_ipod: str = (
        "(iPod; U; CPU iPhone OS {0}_{1} like Mac OS X; {2})"
        " AppleWebKit/{3} (KHTML, like Gecko) Version/{4}.0.5"
        " Mobile/8B{5} Safari/6{6}"
    )
    locale: str = localeLang().replace("_", "-")
    platforms: ElementsType = (
        tmplt_win.format(win_platform_token(), saf, ver, saf),
        tmplt_mac.format(
            mac_platform_token(),
            random.randint(2, 6),
            locale,
            saf,
            ver,
            saf,
        ),
        tmplt_ipod.format(
            random.randint(3, 4),
            random.randint(0, 3),
            locale,
            saf,
            random.randint(3, 4),
            random.randint(111, 119),
            saf,
        ),
    )

    return "Mozilla/5.0 " + random_element(platforms)


def opera() -> str:
    token: str = linux_platform_token() if random.getrandbits(1) else win_platform_token()
    locale: str = localeLang().replace("_", "-")
    platform: str = f"({token}; {locale}) Presto/2.9.{random.randint(160, 190)} " f"Version/{random.randint(10, 12)}.00"
    return f"Opera/{random.randint(8, 9)}.{random.randint(10, 99)}.{platform}"


def internet_explorer() -> str:
    return (
        f"Mozilla/5.0 (compatible; MSIE {random.randint(5, 9)}.0; "
        f"{win_platform_token()}; "
        f"Trident/{random.randint(3, 5)}.{random.randint(0, 1)})"
    )


def random_element(elements: ElementsType = ("a", "b", "c")) -> T:
    return random_elements(elements, length=1)[0]


def win_platform_token() -> str:
    return random_element(win_platform_tokens)


def linux_platform_token() -> str:
    return f"X11; Linux {random_element(linux_processors)}"


def mac_platform_token() -> str:
    return f"Macintosh; {random_element(mac_processors)} Mac OS X 10_" f"{random.randint(5, 12)}_{random.randint(0, 9)}"


def android_platform_token() -> str:
    return f"Android {random_element(android_versions)}"


def ios_platform_token() -> str:
    apple_device: str = random_element(apple_devices)
    ios_version: str = random_element(ios_versions)
    return f"{apple_device}; CPU {apple_device} " f'OS {ios_version.replace(".", "_")} like Mac OS X'



import platform

_python3 = int(platform.python_version_tuple()[0]) >= 3


class MultiKeyDict(object):
    """The purpose of this type is to provide a multi-key dictionary.
    This kind of dictionary has a similar interface to the standard dictionary, and indeed if used
    with single key key elements - it's behaviour is the same as for a standard dict().

    However it also allows for creation of elements using multiple keys (using tuples/lists).
    Such elements can be accessed using either of those keys (e.g read/updated/deleted).
    Dictionary provides also an extended interface for iterating over items and keys by the key type.
    This can be useful e.g.: when creating dictionaries with (index,name) allowing one to iterate over
    items using either: names or indexes. It can be useful for many many other similar use-cases,
    and there is no limit to the number of keys used to map to the value.

    There are also methods to find other keys mapping to the same value as the specified keys etc.
    Refer to examples and test code to see it in action.

    simple example:
        k = MultiKeyDict()
        k[100] = 'hundred'  # add item to the dictionary (as for normal dictionary)

        # but also:
        # below creates entry with two possible key types: int and str,
        # mapping all keys to the assigned value
        k[1000, 'kilo', 'k'] = 'kilo (x1000)'
        print k[1000]   # will print 'kilo (x1000)'
        print k['k']  # will also print 'kilo (x1000)'

        # the same way objects can be updated, and if an object is updated using one key, the new value will
        # be accessible using any other key, e.g. for example above:
        k['kilo'] = 'kilo'
        print k[1000] # will print 'kilo' as value was updated
    """

    def __init__(self, mapping_or_iterable=None, **kwargs):
        """Initializes dictionary from an optional positional argument and a possibly empty set of keyword arguments."""
        self.items_dict = {}
        if mapping_or_iterable is not None:
            if type(mapping_or_iterable) is dict:
                mapping_or_iterable = mapping_or_iterable.items()
            for kv in mapping_or_iterable:
                if len(kv) != 2:
                    raise Exception(
                        "Iterable should contain tuples with exactly two values but specified: {0}.".format(kv),
                    )
                self[kv[0]] = kv[1]
        for keys, value in kwargs.items():
            self[keys] = value

    def __getitem__(self, key):
        """Return the value at index specified as key."""
        return self.items_dict[self.__dict__[str(type(key))][key]]

    def __setitem__(self, keys, value):
        """Set the value at index (or list of indexes) specified as keys.
        Note, that if multiple key list is specified, either:
          - none of keys should map to an existing item already (item creation), or
          - all of keys should map to exactly the same item (as previously created)
              (item update)
        If this is not the case - KeyError is raised."""
        if type(keys) in [tuple, list]:
            num_of_keys_we_have = 0

            for x in keys:
                try:
                    self.__getitem__(x)
                    num_of_keys_we_have += 1
                except Exception:
                    continue

            if num_of_keys_we_have:
                all_select_same_item = True
                direct_key = None
                for key in keys:
                    key_type = str(type(key))
                    try:
                        if not direct_key:
                            direct_key = self.__dict__[key_type][key]
                        else:
                            new = self.__dict__[key_type][key]
                            if new != direct_key:
                                all_select_same_item = False
                                break
                    except Exception:
                        all_select_same_item = False
                        break

                if not all_select_same_item:
                    raise KeyError(", ".join(str(key) for key in keys))

            first_key = keys[0]  # combination if keys is allowed, simply use the first one
        else:
            first_key = keys

        key_type = str(type(first_key))  # find the intermediate dictionary..
        if first_key in self:
            self.items_dict[self.__dict__[key_type][first_key]] = value  # .. and update the object if it exists..
        else:
            if type(keys) not in [tuple, list]:
                key = keys
                keys = [keys]
            self.__add_item(value, keys)  # .. or create it - if it doesn't

    def __delitem__(self, key):
        """Called to implement deletion of self[key]."""
        key_type = str(type(key))

        if key in self and self.items_dict and (self.__dict__[key_type][key] in self.items_dict):
            intermediate_key = self.__dict__[key_type][key]

            # remove the item in main dictionary
            del self.items_dict[intermediate_key]

            # and remove all references (if there were other keys)
            for k in self.get_other_keys(key, True):
                key_type = str(type(k))
                if key_type in self.__dict__ and k in self.__dict__[key_type]:
                    del self.__dict__[key_type][k]

        else:
            raise KeyError(key)

    def __contains__(self, key):
        """Returns True if this object contains an item referenced by the key."""
        key_type = str(type(key))
        if key_type in self.__dict__ and key in self.__dict__[key_type]:
            return True

        return False

    def has_key(self, key):
        """Returns True if this object contains an item referenced by the key."""
        return key in self

    def get_other_keys(self, key, including_current=False):
        """Returns list of other keys that are mapped to the same value as specified key.
        @param key - key for which other keys should be returned.
        @param including_current if set to True - key will also appear on this list."""
        other_keys = []
        if key in self:
            other_keys.extend(self.__dict__[str(type(key))][key])
            if not including_current:
                other_keys.remove(key)
        return other_keys

    def iteritems(self, key_type=None, return_all_keys=False):
        """Returns an iterator over the dictionary's (key, value) pairs.
        @param key_type if specified, iterator will be returning only (key,value) pairs for this type of key.
               Otherwise (if not specified) ((keys,...), value)
               i.e. (tuple of keys, values) pairs for all items in this dictionary will be generated.
        @param return_all_keys if set to True - tuple of keys is retuned instead of a key of this type."""

        if key_type is None:
            for item in self.items_dict.items():
                yield item
            return
        used_keys = set()
        key = str(key_type)
        if key in self.__dict__:
            for key, keys in self.__dict__[key].items():
                if keys in used_keys:
                    continue
                used_keys.add(keys)
                value = self.items_dict[keys]
                if not return_all_keys:
                    keys = tuple(k for k in keys if isinstance(k, key_type))
                yield keys, value

    def iterkeys(self, key_type=None, return_all_keys=False):
        """Returns an iterator over the dictionary's keys.
        @param key_type if specified, iterator for a dictionary of this type will be used.
               Otherwise (if not specified) tuples containing all (multiple) keys
               for this dictionary will be generated.
        @param return_all_keys if set to True - tuple of keys is retuned instead of a key of this type."""
        if key_type is not None:
            the_key = str(key_type)
            if the_key in self.__dict__:
                for key in self.__dict__[the_key]:
                    if return_all_keys:
                        yield self.__dict__[the_key][key]
                    else:
                        yield key
        else:
            for keys in self.items_dict:
                yield keys

    def itervalues(self, key_type=None):
        """Returns an iterator over the dictionary's values.
        @param key_type if specified, iterator will be returning only values pointed by keys of this type.
               Otherwise (if not specified) all values in this dictinary will be generated."""
        if key_type is not None:
            intermediate_key = str(key_type)
            if intermediate_key in self.__dict__:
                for direct_key in self.__dict__[intermediate_key].values():
                    yield self.items_dict[direct_key]
        else:
            for value in self.items_dict.values():
                yield value

    if _python3:
        items = iteritems
    else:

        def items(self, key_type=None, return_all_keys=False):
            return list(self.iteritems(key_type, return_all_keys))

        items.__doc__ = iteritems.__doc__

    def keys(self, key_type=None):
        """Returns a copy of the dictionary's keys.
        @param key_type if specified, only keys for this type will be returned.
             Otherwise list of tuples containing all (multiple) keys will be returned."""
        if key_type is not None:
            intermediate_key = str(key_type)
            if intermediate_key in self.__dict__:
                return self.__dict__[intermediate_key].keys()
            return None
        else:
            all_keys = {}  # in order to preserve keys() type (dict_keys for python3)
            for keys in self.items_dict:
                all_keys[keys] = None
            return all_keys.keys()

    def values(self, key_type=None):
        """Returns a copy of the dictionary's values.
        @param key_type if specified, only values pointed by keys of this type will be returned.
             Otherwise list of all values contained in this dictionary will be returned."""
        if key_type is not None:
            all_items = {}  # in order to preserve keys() type (dict_values for python3)
            keys_used = set()
            direct_key = str(key_type)
            if direct_key in self.__dict__:
                for intermediate_key in self.__dict__[direct_key].values():
                    if intermediate_key not in keys_used:
                        all_items[intermediate_key] = self.items_dict[intermediate_key]
                        keys_used.add(intermediate_key)
            return all_items.values()
        else:
            return self.items_dict.values()

    def __len__(self):
        """Returns number of objects in dictionary."""
        length = 0
        if "items_dict" in self.__dict__:
            length = len(self.items_dict)
        return length

    def __add_item(self, item, keys=None):
        """Internal method to add an item to the multi-key dictionary"""
        if not keys or not len(keys):
            raise Exception(
                "Error in %s.__add_item(%s, keys=tuple/list of items): need to specify a tuple/list containing at least one key!"
                % (self.__class__.__name__, str(item)),
            )
        direct_key = tuple(keys)  # put all keys in a tuple, and use it as a key
        for key in keys:
            key_type = str(type(key))

            # store direct key as a value in an intermediate dictionary
            if key_type not in self.__dict__:
                self.__setattr__(key_type, {})
            self.__dict__[key_type][key] = direct_key

            # store the value in the actual dictionary
            if "items_dict" not in self.__dict__:
                self.items_dict = {}
            self.items_dict[direct_key] = item

    def get(self, key, default=None):
        """Return the value at index specified as key."""
        if key in self:
            return self.items_dict[self.__dict__[str(type(key))][key]]
        else:
            return default

    def __str__(self):
        items = []

        def str_repr(x):
            return "'%s'" % x if type(x) == str else str(x)

        if hasattr(self, "items_dict"):
            for keys, value in self.items():
                keys_str = [str_repr(k) for k in keys]
                items.append("(%s): %s" % (", ".join(keys_str), str_repr(value)))
        dict_str = "{%s}" % (", ".join(items))
        return dict_str


def test_MultiKeyDict():
    def contains_all(cont, in_items):
        return False not in [(c in cont) for c in in_items]

    m = MultiKeyDict()
    assert len(m) == 0, "expected len(m) == 0"
    all_keys = []

    m["aa", 12, 32, "mmm"] = 123  # create a value with multiple keys..
    assert len(m) == 1, "expected len(m) == 1"
    all_keys.append(("aa", "mmm", 32, 12))  # store it for later

    # try retrieving other keys mapped to the same value using one of them
    res = m.get_other_keys("aa")
    expected = ["mmm", 32, 12]
    assert set(res) == set(expected), "get_other_keys('aa'): {0} other than expected: {1} ".format(res, expected)

    # try retrieving other keys mapped to the same value using one of them: also include this key
    res = m.get_other_keys(32, True)
    expected = ["aa", "mmm", 32, 12]
    assert set(res) == set(expected), "get_other_keys(32): {0} other than expected: {1} ".format(res, expected)

    assert m.has_key("aa") is True, "expected m.has_key('aa') == True"
    assert m.has_key("aab") is False, "expected m.has_key('aab') == False"

    assert m.has_key(12) is True, "expected m.has_key(12) == True"
    assert m.has_key(13) is False, "expected m.has_key(13) == False"
    assert m.has_key(32) is True, "expected m.has_key(32) == True"

    m["something else"] = "abcd"
    assert len(m) == 2, "expected len(m) == 2"
    all_keys.append(("something else",))  # store for later

    m[23] = 0
    assert len(m) == 3, "expected len(m) == 3"
    all_keys.append((23,))  # store for later

    # check if it's possible to read this value back using either of keys
    assert m["aa"] == 123, "expected m['aa'] == 123"
    assert m[12] == 123, "expected m[12] == 123"
    assert m[32] == 123, "expected m[32] == 123"
    assert m["mmm"] == 123, "expected m['mmm'] == 123"

    # now update value and again - confirm it back - using different keys..
    m["aa"] = 45
    assert m["aa"] == 45, "expected m['aa'] == 45"
    assert m[12] == 45, "expected m[12] == 45"
    assert m[32] == 45, "expected m[32] == 45"
    assert m["mmm"] == 45, "expected m['mmm'] == 45"

    m[12] = "4"
    assert m["aa"] == "4", "expected m['aa'] == '4'"
    assert m[12] == "4", "expected m[12] == '4'"

    # test __str__
    m_str_exp = "{(23): 0, ('aa', 'mmm', 32, 12): '4', ('something else'): 'abcd'}"
    m_str = str(m)
    assert len(m_str) > 0, "str(m) should not be empty!"
    assert m_str[0] == "{", "str(m) should start with '{', but does with '%c'" % m_str[0]
    assert m_str[-1] == "}", "str(m) should end with '}', but does with '%c'" % m_str[-1]

    # check if all key-values are there as expected. They might be sorted differently
    def get_values_from_str(dict_str):
        sorted_keys_and_values = []
        for k in dict_str.split(", ("):
            keys, val = k.strip("{}() ").replace(")", "").split(":")
            keys = tuple(sorted([k.strip() for k in keys.split(",")]))
            sorted_keys_and_values.append((keys, val))
        return sorted_keys_and_values

    exp = get_values_from_str(m_str_exp)
    act = get_values_from_str(m_str)
    assert set(act) == set(exp), "str(m) values: '{0}' are not {1} ".format(act, exp)

    # try accessing / creating new (keys)-> value mapping whilst one of these
    # keys already maps to a value in this dictionarys
    try:
        m["aa", "bb"] = "something new"
        raise AssertionError("Should not allow adding multiple-keys when one of keys ('aa') already exists!")
    except KeyError:
        pass

    # now check if we can get all possible keys (formed in a list of tuples)
    # each tuple containing all keys)
    res = sorted([sorted([str(x) for x in k]) for k in m])
    expected = sorted([sorted([str(x) for x in k]) for k in all_keys])
    assert res == expected, "unexpected values from m.keys(), got:\n%s\n expected:\n%s" % (res, expected)

    # check default items (which will unpack tupe with key(s) and value)
    num_of_elements = 0
    for keys, value in m.items():
        sorted_keys = sorted([str(k) for k in keys])
        num_of_elements += 1
        assert sorted_keys in expected, "m.items(): unexpected keys: %s" % (sorted_keys)
        assert m[keys[0]] == value, "m.items(): unexpected value: %s (keys: %s)" % (
            value,
            keys,
        )
    assert num_of_elements > 0, "m.items() returned generator that did not produce anything"

    # test default iterkeys()
    num_of_elements = 0
    for keys in m:
        num_of_elements += 1
        keys_s = sorted([str(k) for k in keys])
        assert keys_s in expected, "m.keys(): unexpected keys: {0}".format(keys_s)

    assert num_of_elements > 0, "m.iterkeys() returned generator that did not produce anything"

    # test iterkeys(int, True): useful to get all info from the dictionary
    # dictionary is iterated over the type specified, but all keys are returned.
    num_of_elements = 0
    for keys in m.iterkeys(int, True):
        keys_s = sorted([str(k) for k in keys])
        num_of_elements += 1
        assert keys_s in expected, "m.iterkeys(int, True): unexpected keys: {0}".format(keys_s)
    assert num_of_elements > 0, "m.iterkeys(int, True) returned generator that did not produce anything"

    # test values for different types of keys()
    expected = {0, "4"}
    res = set(m.values(int))
    assert res == expected, "m.values(int) are {0}, but expected: {1}.".format(res, expected)

    expected = sorted(["4", "abcd"])
    res = sorted(m.values(str))
    assert res == expected, "m.values(str) are {0}, but expected: {1}.".format(res, expected)

    current_values = {0, "4", "abcd"}  # default (should give all values)
    res = set(m.values())
    assert res == current_values, "m.values() are {0}, but expected: {1}.".format(res, current_values)

    # test itervalues() (default) - should return all values. (Itervalues for other types are tested below)
    vals = set()
    for value in m.itervalues():
        vals.add(value)
    assert current_values == vals, "itervalues(): expected {0}, but collected {1}".format(current_values, vals)

    # test items(int)
    items_for_int = sorted([((12, 32), "4"), ((23,), 0)])
    assert items_for_int == sorted(m.items(int)), "items(int): expected {0}, but collected {1}".format(
        items_for_int,
        sorted(m.items(int)),
    )

    # test items(str)
    items_for_str = {(("aa", "mmm"), "4"), (("something else",), "abcd")}
    res = set(m.items(str))
    assert set(res) == items_for_str, "items(str): expected {0}, but collected {1}".format(items_for_str, res)

    # test items() (default - all items)
    # we tested keys(), values(), and __get_item__ above so here we'll re-create all_items using that
    all_items = set()
    keys = m.keys()
    m.values()
    for k in keys:
        all_items.add((tuple(k), m[k[0]]))

    res = set(m.items())
    assert all_items == res, "items() (all items): expected {0},\n\t\t\t\tbut collected {1}".format(all_items, res)

    # now test deletion..
    curr_len = len(m)
    del m[12]
    assert len(m) == curr_len - 1, "expected len(m) == %d" % (curr_len - 1)
    assert not m.has_key(12), "expected deleted key to no longer be found!"

    # try again
    try:
        del m["aa"]
        raise AssertionError("cant remove again: item m['aa'] should not exist!")
    except KeyError:
        pass

    # try to access non-existing
    try:
        k = m["aa"]
        raise AssertionError("removed item m['aa'] should not exist!")
    except KeyError:
        pass

    # try to access non-existing with a different key
    try:
        k = m[12]
        raise AssertionError("removed item m[12] should not exist!")
    except KeyError:
        pass

    # prepare for other tests (also testing creation of new items)
    del m
    m = MultiKeyDict()
    tst_range = list(range(10, 40)) + list(range(50, 70))
    for i in tst_range:
        m[i] = i  # will create a dictionary, where keys are same as items

    # test items()
    for key, value in m.items(int):
        assert key == (value,), "items(int): expected {0}, but received {1}".format(key, value)

    # test iterkeys()
    num_of_elements = 0
    returned_keys = set()
    for key in m.iterkeys(int):
        returned_keys.add(key)
        num_of_elements += 1
    assert num_of_elements > 0, "m.iteritems(int) returned generator that did not produce anything"
    assert returned_keys == set(tst_range), "iterkeys(int): expected {0}, but received {1}".format(expected, key)

    # test itervalues(int)
    num_of_elements = 0
    returned_values = set()
    for value in m.itervalues(int):
        returned_values.add(value)
        num_of_elements += 1
    assert num_of_elements > 0, "m.itervalues(int) returned generator that did not produce anything"
    assert returned_values == set(tst_range), "itervalues(int): expected {0}, but received {1}".format(expected, value)

    # test values(int)
    res = sorted(m.values(int))
    assert res == tst_range, "m.values(int) is not as expected."

    # test keys()
    assert set(m.keys(int)) == set(tst_range), "m.keys(int) is not as expected."

    # test setitem with multiple keys
    m["xy", 999, "abcd"] = "teststr"
    try:
        m["xy", 998] = "otherstr"
        raise AssertionError("creating / updating m['xy', 998] should fail!")
    except KeyError:
        pass

    # test setitem with multiple keys
    m["cd"] = "somethingelse"
    try:
        m["cd", 999] = "otherstr"
        raise AssertionError("creating / updating m['cd', 999] should fail!")
    except KeyError:
        pass

    m["xy", 999] = "otherstr"
    assert m["xy"] == "otherstr", "m['xy'] is not as expected."
    assert m[999] == "otherstr", "m[999] is not as expected."
    assert m["abcd"] == "otherstr", "m['abcd'] is not as expected."

    m["abcd", "xy"] = "another"
    assert m["xy"] == "another", "m['xy'] is not == 'another'."
    assert m[999] == "another", "m[999] is not == 'another'"
    assert m["abcd"] == "another", "m['abcd'] is not  == 'another'."

    # test get functionality of basic dictionaries
    m["CanIGet"] = "yes"
    assert m.get("CanIGet") == "yes"
    assert m.get("ICantGet") is None
    assert m.get("ICantGet", "Ok") == "Ok"

    k = MultiKeyDict()
    k["1:12", 1] = "key_has_:"
    k.items()  # should not cause any problems to have : in key
    assert k[1] == "key_has_:", "k[1] is not equal to 'abc:def:ghi'"

    import datetime

    n = datetime.datetime.now()
    l = MultiKeyDict()
    l[n] = "now"  # use datetime obj as a key

    # test keys..
    res = list(l.keys())[0]  # for python3 keys() returns dict_keys dictionarly
    expected = (n,)
    assert expected == res, 'Expected "{0}", but got: "{1}"'.format(expected, res)

    res = list(l.keys(datetime.datetime))[0]
    assert n == res, "Expected {0} as a key, but got: {1}".format(n, res)

    res = list(l.values())  # for python3 keys() returns dict_values dictionarly
    expected = ["now"]
    assert res == expected, "Expected values: {0}, but got: {1}".format(expected, res)

    # test items..
    exp_items = [((n,), "now")]
    r = list(l.items())
    assert r == exp_items, "Expected for items(): tuple of keys: {0}, but got: {1}".format(r, exp_items)
    assert exp_items[0][1] == "now", "Expected for items(): value: {0}, but got: {1}".format("now", exp_items[0][1])

    x = MultiKeyDict({("k", "kilo"): 1000, ("M", "MEGA", 1000000): 1000000}, milli=0.01)
    assert x["k"] == 1000, "x['k'] is not equal to 1000"
    x["kilo"] = "kilo"
    assert x["kilo"] == "kilo", "x['kilo'] is not equal to 'kilo'"

    y = MultiKeyDict([(("two", "duo"), 2), (("one", "uno"), 1), ("three", 3)])

    assert y["two"] == 2, "y['two'] is not equal to 2"
    y["one"] = "one"
    assert y["one"] == "one", "y['one'] is not equal to 'one'"

    try:
        y = MultiKeyDict([(("two", "duo"), 2), ("one", "uno", 1), ("three", 3)])
        raise AssertionError("creating dictionary using iterable with tuples of size > 2 should fail!")
    except:
        pass

    print("All test passed OK!")


__all__ = ["MultiKeyDict"]

if __name__ == "__main__":
    try:
        test_MultiKeyDict()
    except KeyboardInterrupt:
        print("\n(interrupted by user)")



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  常用装饰器
"""

import json
import logging
import sys
import threading
import time
import traceback
from functools import wraps

logger = logging.getLogger(__name__)


class ExcContextManager:
    def __init__(self, verbose=100, nr_exc=True):
        """
        用上下文管理器捕获异常,可对代码片段进行错误捕捉,比装饰器更细腻
        Args:
            verbose: 打印错误的深度,对应traceback对象的limit,为正整数
            nr_exc:
        """
        self._verbose = verbose
        self._nr_exc = nr_exc

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        exc_str = str(exc_type) + "  :  " + str(exc_val)
        exc_str_color = "\033[0;30;45m%s\033[0m" % exc_str
        if self._nr_exc and exc_tb is not None:
            msg = "\n".join(traceback.format_tb(exc_tb)[: self._verbose]) + exc_str_color
            print(msg)
        return self._nr_exc  # __exit__方法必须return True才会不重新抛出错误


def handle_exception(retry_times=0, error_detail_level=0, is_throw_error=False, time_sleep=0, opt_logger=False):
    """
    捕获函数错误的装饰器,重试并打印日志
    Args:
        retry_times: 重试次数
        error_detail_level: 为0打印exception提示,为1打印3层深度的错误堆栈,为2打印所有深度层次的错误堆栈
        is_throw_error: 在达到最大次数时候是否重新抛出错误
        time_sleep: 休眠时间
        opt_logger: 是否打印日志
    Returns:
    Examples:
        >>> from custard.function import bind_run_many_times
        >>> import json
        >>> import time
        >>> @bind_run_many_times(3)
        ... @handle_exception(2, 1)
        ... def test_superposition():
        ...    json.loads('a', ac='ds')
        >>> test_superposition()

        >>> @handle_exception(2)
        ... def test_handle_exception():
        ...     pass
        >>> test_handle_exception()
    """
    if error_detail_level not in [0, 1, 2]:
        detail = "error_detail_level参数必须设置为0 、1 、2"
        raise Exception(detail)

    def _handle_exception(func):
        @wraps(func)
        def __handle_exception(*args, **kwargs):
            for i in range(0, retry_times + 1):
                try:
                    result = func(*args, **kwargs)
                    if i:
                        info_msg = f'{"# " * 40}\n调用成功,调用方法--> [  {func.__name__}  ] 第  {i + 1}  次重试成功'
                        logger.info(opt_logger, info_msg)
                    return result

                except Exception as e:
                    error_info = ""
                    if error_detail_level == 0:
                        error_info = "错误类型是:" + str(e.__class__) + "  " + str(e)
                    elif error_detail_level == 1:
                        error_info = "错误类型是:" + str(e.__class__) + "  " + traceback.format_exc(limit=3)
                    elif error_detail_level == 2:
                        error_info = "错误类型是:" + str(e.__class__) + "  " + traceback.format_exc()
                    err_msg = f'{"-" * 40}\n记录错误日志,调用方法--> [  {func.__name__}  ] 第  {i + 1}  次错误重试, {error_info}'
                    if opt_logger:
                        logger.error(err_msg)
                    if i == retry_times and is_throw_error:  # 达到最大错误次数后,重新抛出错误
                        raise e
                time.sleep(time_sleep)
            return None

        return __handle_exception

    return _handle_exception


def keep_circulating(
    time_sleep=0.001,
    exit_if_function_run_succeed=True,
    is_display_detail_exception=True,
    block=True,
    daemon=False,
    opt_logger=False,
):
    """
    间隔一段时间,一直循环运行某个方法的装饰器
    Args:
        time_sleep: 循环的间隔时间
        exit_if_function_run_succeed: 如果成功了就退出循环
        is_display_detail_exception:  是否显示细节异常
        block: 是否阻塞主主线程,False时候开启一个新的线程运行while 1。
        daemon: 如果使用线程,那么是否使用守护线程,使这个while 1有机会自动结束。
        opt_logger: 是否打印日志
    Returns:
    Examples:
        >>> @keep_circulating(3, block=False)
        ... def test_keep_circulating(index = 0):
        ...     if isinstance(index, str):
        ...        pass
        ...     else:
        ...         for index in range(5):
        ...             if index <2:
        ...                 print("每隔3秒,一直打印   " + time.strftime('%H:%M:%S'))
        ...             else:
        ...                 test_keep_circulating(index="test")
        ...             index += 1
        >>> test_keep_circulating()
    """

    def _keep_circulating(func):
        @wraps(func)
        def __keep_circulating(*args, **kwargs):
            def ___keep_circulating():
                while 1:
                    try:
                        result = func(*args, **kwargs)
                        if exit_if_function_run_succeed:
                            return result
                    except Exception as e:
                        msg = (
                            func.__name__ + "   运行出错\n " + traceback.format_exc(limit=10)
                            if is_display_detail_exception
                            else str(e)
                        )
                        if opt_logger:
                            logger.info(msg)
                    finally:
                        time.sleep(time_sleep)
                return None

            if block:
                return ___keep_circulating()
            else:
                threading.Thread(target=___keep_circulating, daemon=daemon).start()
                return None

        return __keep_circulating

    return _keep_circulating


def where_is_it_called(func):
    """
    记录函数被什么文件的哪一行代码所调用,非常犀利黑科技的装饰器
    """

    @wraps(func)
    def _where_is_it_called(*args, **kwargs):
        # 获取被调用函数名称
        func_name = func.__name__
        # 什么函数调用了此函数
        which_fun_call_this = sys._getframe(1).f_code.co_name  # NOQA

        # 获取被调用函数在被调用时所处代码行数
        line = sys._getframe().f_back.f_lineno

        # 获取被调用函数所在模块文件名
        file_name = sys._getframe(1).f_code.co_filename

        msg = (
            f"文件[{func.__code__.co_filename}]的第[{func.__code__.co_firstlineno}]行"
            f"即模块 [{func.__module__}] 中的方法 [{func_name}] "
            f"正在被文件 [{file_name}] 中的方法 [{which_fun_call_this}] 中的"
            f"第 [{line}] 行处调用,传入的参数为[{args},{kwargs}]"
        )
        try:
            start_time = time.time()
            result = func(*args, **kwargs)
            result_raw = result
            t_spend = round(time.time() - start_time, 2)
            if isinstance(result, dict):
                result = json.dumps(result)
            if len(str(result)) > 200:
                result = str(result)[0:200]
            msg = "执行函数[{}]消耗的时间是{}秒,返回的结果是 --> ".format(func_name, t_spend) + str(result)
            print(msg)
            return result_raw
        except Exception as e:
            debug_msg = "执行函数{},发生错误".format(func_name)
            print(debug_msg, e)
            raise e

    return _where_is_it_called


def bind_run_many_times(times=1, opt_logger=False):
    """
    设置函数运行次数
    Args:
        times: 运行次数
        opt_logger: 日志开关
    Returns: 没有捕获错误,出错误就中断运行,可以配合handle_exception装饰器不管是否错误都运行n次。
    Examples:
        >>> import time
        >>> @bind_run_many_times(5)
        ... def test_run_many_times_():
        ...     print('hello world!')
        ...     time.sleep(0.2)
        >>> test_run_many_times_()
    """

    def _run_many_times(func):
        @wraps(func)
        def __run_many_times(*args, **kwargs):
            for i in range(times):
                msg = "当前是第 {} 次运行[ {} ]函数".format(i + 1, func.__name__)
                if opt_logger:
                    logger.info(msg)
                func(*args, **kwargs)

        return __run_many_times

    return _run_many_times


class BatchTask:
    @staticmethod
    def list_jobs(func_name, context):
        """
        批量任务处理
        Args:
            func_name:
            context:
        Returns:
        Examples:
            >>> examples = [False, True, 1]
            >>> def test_status(**kwargs):
            ...     return kwargs["example"] is True
            >>> def test_list_jobs(example):
            ...     return  test_status(example=example)
            >>> BatchTask.list_jobs(test_list_jobs, examples)
        """
        accord, no_accord = [], []
        if not isinstance(context, list):
            context = [context]
        for index in context:
            if func_name.__call__(index):
                accord.append(index)
            else:
                no_accord.append(index)
        return True if no_accord == [] else (accord, no_accord)


def sync_redis_lock(key, client, lock_time_out, opt_logger=False):
    """
    redis分布式锁装饰器
    todo:锁的超时时间需要考虑任务函数执行时间,
    任务函数执行时间可能会越来越长,
    如果函数执行时间超出锁设置时间可能会存在执行多次情况
    redis命令
        以redis的key为锁名,value为超时时间
        get 获取对应key的value
        setnx 设置一个str类型.key存在不执行,返回0(受影响行数).key不存在插入一条数据,返回1
        gest 为一个str类型设置一个新的value并返回之前的value
    原理:
        设置锁:任务函数执行之前先设置锁,并添加超时时间(当前时间戳+超时时间),需要把任务执行时间考虑进去
        设置锁失败:查询锁的value是否小于当前时间,小于(锁还没超时),大于(锁超时)
        锁超时:(不可直接删除锁,如果多个进程都删除了锁,会同时获得锁),gest设置锁的新值(可能多个进程设置锁的新值,但只有一进程gest的返回值,小于当前时间戳,此进程才能执行任务)
    Args:
        key:
        client:
        lock_time_out:
        opt_logger:

    Returns:
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # 当前时间
            now_time = int(time.time())
            # 获取并设置锁(value为当前时间+超时时间) 成功返回1,失败返回0
            lock = client.setnx(key, now_time + lock_time_out)
            if opt_logger:
                logger.info("Now Time: %s" % now_time, "lock: %s" % lock)
            if lock == 1:
                func(*args, **kwargs)
                if opt_logger:
                    logger.info("创建锁任务结束")
                client.delete(key)
                if opt_logger:
                    logger.info("删除锁任务结束")
            else:
                # 锁存在,查看锁是否超时
                lock_set_time = int(client.get(key))
                # 锁超时
                if now_time > lock_set_time:
                    # 重新设置锁(不能直接删除,可能出现多个进程同时检查到锁超时都获取到锁的情况)
                    # gest设置新值返回旧值
                    old_lock_time = int(client.gest(key, now_time + lock_time_out))
                    if opt_logger:
                        logger.info("old_lock_time", type(old_lock_time), old_lock_time)
                    # 如果目前时间小于旧锁时间,表示有别的进程获取到了锁.放弃执行任务,反之执行任务
                    if now_time > old_lock_time:
                        if opt_logger:
                            logger.info("锁超时开始任务")
                        func(*args, **kwargs)
                        if opt_logger:
                            logger.info("锁超时结束任务")
                        client.delete(key)
                        if opt_logger:
                            logger.info("锁超时删除锁")

            if opt_logger:
                logger.info("没有执行任务")

        return wrapper

    return decorator


def synchronized_lock(func):
    """
    线程锁装饰器,可以加在单例模式上
    Args:
        func:
    Returns:
    """
    func.__lock__ = threading.Lock()

    @wraps(func)
    def lock_func(*args, **kwargs):
        with func.__lock__:
            return func(*args, **kwargs)

    return lock_func


def singleton_lock(cls):
    """
    单例模式装饰器,新加入线程锁,更牢固的单例模式
    主要解决多线程如100线程同时实例化情况下可能会出现三例四例的情况,实测。
    Args:
        cls:
    Returns:
    """
    _instance = {}
    singleton_lock.__lock = threading.Lock()

    @wraps(cls)
    def _singleton(*args, **kwargs):
        with singleton_lock.__lock:
            if cls not in _instance:
                _instance[cls] = cls(*args, **kwargs)
            return _instance[cls]

    return _singleton



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  paginator.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from typing import Callable, Optional, Sequence, TypeVar

from .api import create_page, resolve_params
from .bases import AbstractPage, AbstractParams

T = TypeVar("T")


def paginate(
    sequence: Sequence[T],
    params: Optional[AbstractParams] = None,
    length_function: Callable[[Sequence[T]], int] = len,
) -> AbstractPage[T]:
    params = resolve_params(params)
    raw_params = params.to_raw_params()

    return create_page(
        items=sequence[raw_params.offset : raw_params.offset + raw_params.limit],
        total=length_function(sequence),
        params=params,
    )


__all__ = ["paginate"]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from .api import (
    add_pagination,
    create_page,
    request,
    resolve_params,
    response,
    set_page,
)
from .limit_offset import LimitOffsetPage, LimitOffsetParams
from .pagination import Page, Params
from .paginator import paginate

__all__ = [
    "add_pagination",
    "create_page",
    "request",
    "resolve_params",
    "response",
    "set_page",
    "Page",
    "Params",
    "LimitOffsetPage",
    "LimitOffsetParams",
    "paginate",
]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  mini_racer.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""

import ctypes
import datetime
import json
import os
import sys
import sysconfig
import threading

try:
    import pkg_resources
except ImportError:
    pkg_resources = None  # pragma: no cover


def _get_libc_name():
    """Return the libc name of the current system."""
    target = sysconfig.get_config_var("HOST_GNU_TYPE")
    if target is not None and target.endswith("musl"):
        return "muslc"
    return "glibc"


def _get_lib_path(name):
    """Return the path of the library called `name` on the current system."""
    if os.name == "posix" and sys.platform == "darwin":
        prefix, ext = "lib", ".dylib"
    elif sys.platform == "win32":
        prefix, ext = "", ".dll"
    else:
        prefix, ext = "lib", ".{}.so".format(_get_libc_name())
    fn = None
    meipass = getattr(sys, "_MEIPASS", None)
    if meipass is not None:
        fn = os.path.join(meipass, prefix + name + ext)
    if fn is None and pkg_resources is not None:
        fn = pkg_resources.resource_filename("py_mini_racer", prefix + name + ext)
    if fn is None:
        root_dir = os.path.dirname(os.path.abspath(__file__))
        fn = os.path.join(root_dir, prefix + name + ext)
    return fn


# In python 3 the extension file name depends on the python version
EXTENSION_PATH = _get_lib_path("mini_racer")
EXTENSION_NAME = os.path.basename(EXTENSION_PATH) if EXTENSION_PATH is not None else None

if sys.version_info[0] < 3:
    UNICODE_TYPE = unicode  # noqa: F821
else:
    UNICODE_TYPE = str


class MiniRacerBaseException(Exception):
    """Base MiniRacer exception."""


class JSParseException(MiniRacerBaseException):
    """JavaScript could not be parsed."""


class JSEvalException(MiniRacerBaseException):
    """JavaScript could not be executed."""


class JSOOMException(JSEvalException):
    """JavaScript execution ran out of memory."""


class JSTimeoutException(JSEvalException):
    """JavaScript execution timed out."""


class JSConversionException(MiniRacerBaseException):
    """JavaScript type could not be converted."""


class WrongReturnTypeException(MiniRacerBaseException):
    """Invalid type returned by the JavaScript runtime."""


class JSObject(object):
    """JavaScript object."""

    def __init__(self, id):
        self.id = id

    def __hash__(self):
        return self.id


class JSFunction(object):
    """JavaScript function."""


class JSSymbol(object):
    """JavaScript symbol."""


def is_unicode(value):
    """Check if a value is a valid unicode string.

    >>> is_unicode(u'foo')
    True
    >>> is_unicode(u'✌')
    True
    >>> is_unicode(b'foo')
    False
    >>> is_unicode(42)
    False
    >>> is_unicode(('abc',))
    False
    """
    return isinstance(value, UNICODE_TYPE)


def _build_ext_handle():
    if EXTENSION_PATH is None or not os.path.exists(EXTENSION_PATH):
        raise RuntimeError("Native library not available at {}".format(EXTENSION_PATH))

    _ext_handle = ctypes.CDLL(EXTENSION_PATH)

    _ext_handle.mr_init_context.argtypes = [ctypes.c_char_p]
    _ext_handle.mr_init_context.restype = ctypes.c_void_p

    _ext_handle.mr_eval_context.argtypes = [
        ctypes.c_void_p,
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.c_ulong,
        ctypes.c_size_t,
    ]
    _ext_handle.mr_eval_context.restype = ctypes.POINTER(MiniRacerValueStruct)

    _ext_handle.mr_free_value.argtypes = [ctypes.c_void_p, ctypes.c_void_p]

    _ext_handle.mr_free_context.argtypes = [ctypes.c_void_p]

    _ext_handle.mr_heap_stats.argtypes = [ctypes.c_void_p]
    _ext_handle.mr_heap_stats.restype = ctypes.POINTER(MiniRacerValueStruct)

    _ext_handle.mr_low_memory_notification.argtypes = [ctypes.c_void_p]

    _ext_handle.mr_heap_snapshot.argtypes = [ctypes.c_void_p]
    _ext_handle.mr_heap_snapshot.restype = ctypes.POINTER(MiniRacerValueStruct)

    _ext_handle.mr_set_soft_memory_limit.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
    _ext_handle.mr_set_soft_memory_limit.restype = None

    _ext_handle.mr_soft_memory_limit_reached.argtypes = [ctypes.c_void_p]
    _ext_handle.mr_soft_memory_limit_reached.restype = ctypes.c_bool

    _ext_handle.mr_v8_version.restype = ctypes.c_char_p

    return _ext_handle


class MiniRacer(object):
    """
    MiniRacer evaluates JavaScript code using a V8 isolate.

    :cvar json_impl: JSON module used by helper methods default is :py:mod:`json`
    :cvar v8_flags: Flags used for V8 initialization
    :vartype v8_flags: class attribute list of str
    """

    json_impl = json
    v8_flags = ["--single-threaded"]
    ext = None

    def __init__(self):
        if self.__class__.ext is None:
            self.__class__.ext = _build_ext_handle()

        self.ctx = self.ext.mr_init_context(" ".join(self.v8_flags).encode("utf-8"))
        self.lock = threading.Lock()

    @property
    def v8_version(self):
        """Return the V8 version string."""
        return UNICODE_TYPE(self.ext.mr_v8_version())

    def eval(self, code, timeout=None, max_memory=None):
        # type: (str, Optional[int], Optional[int]) -> Any
        """Evaluate JavaScript code in the V8 isolate.

        Side effects from the JavaScript evaluation is persisted inside a context
        (meaning variables set are kept for the next evaluations).

        The JavaScript value returned by the last expression in `code` is converted
        to a Python value and returned by this method. Only primitive types are
        supported (numbers, strings, buffers...). Use the :py:meth:`.execute` method to return
        more complex types such as arrays or objects.

        The evaluation can be interrupted by an exception for several reasons: a limit
        was reached, the code could not be parsed, a returned value could not be
        converted to a Python value.

        :param code: JavaScript code
        :param timeout: number of milliseconds after which the execution is interrupted
        :param max_memory: hard memory limit after which the execution is interrupted
        """

        if is_unicode(code):
            code = code.encode("utf8")

        with self.lock:
            res = self.ext.mr_eval_context(
                self.ctx,
                code,
                len(code),
                ctypes.c_ulong(timeout or 0),
                ctypes.c_size_t(max_memory or 0),
            )
        if not res:
            raise JSConversionException

        return MiniRacerValue(self, res).to_python()

    def execute(self, expr, timeout=None, max_memory=None):
        # type: (str, Optional[int], Optional[int]) -> Any
        """Helper to evaluate a JavaScript expression and return composite types.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using :py:attr:`.json_impl`.

        :param expr: JavaScript expression
        :param timeout: number of milliseconds after which the execution is interrupted
        :param max_memory: hard memory limit after which the execution is interrupted
        """
        wrapped_expr = "JSON.stringify((function(){return (%s)})())" % expr
        ret = self.eval(wrapped_expr, timeout=timeout, max_memory=max_memory)
        if not is_unicode(ret):
            raise ValueError("Unexpected return value type {}".format(type(ret)))
        return self.json_impl.loads(ret)

    def call(self, expr, *args, **kwargs):
        """Helper to call a JavaScript function and return compositve types.

        The `expr` argument refers to a JavaScript function in the current V8
        isolate context. Further positional arguments are serialized using the JSON
        implementation :py:attr:`.json_impl` and passed to the JavaScript function
        as arguments.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using :py:attr:`.json_impl`.

        :param str expr: JavaScript expression referring to a function
        :param encoder: Custom JSON encoder
        :type encoder: JSONEncoder or None
        :param int timeout: number of milliseconds after which the execution is interrupted
        :param int max_memory: hard memory limit after which the execution is interrupted
        """

        encoder = kwargs.get("encoder", None)
        timeout = kwargs.get("timeout", None)
        max_memory = kwargs.get("max_memory", None)

        json_args = self.json_impl.dumps(args, separators=(",", ":"), cls=encoder)
        js = "{expr}.apply(this, {json_args})".format(expr=expr, json_args=json_args)
        return self.execute(js, timeout=timeout, max_memory=max_memory)

    def set_soft_memory_limit(self, limit):
        # type: (int) -> None
        """Set a soft memory limit on this V8 isolate.

        The Garbage Collection will use a more aggressive strategy when
        the soft limit is reached but the execution will not be stopped.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """
        self.ext.mr_set_soft_memory_limit(self.ctx, limit)

    def was_soft_memory_limit_reached(self):
        # type: () -> bool
        """Return true if the soft memory limit was reached on the V8 isolate."""
        return self.ext.mr_soft_memory_limit_reached(self.ctx)

    def low_memory_notification(self):
        """Ask the V8 isolate to collect memory more aggressively."""
        self.ext.mr_low_memory_notification(self.ctx)

    def heap_stats(self):
        """Return the V8 isolate heap statistics."""

        with self.lock:
            res = self.ext.mr_heap_stats(self.ctx)

        if not res:
            return {
                "total_physical_size": 0,
                "used_heap_size": 0,
                "total_heap_size": 0,
                "total_heap_size_executable": 0,
                "heap_size_limit": 0,
            }

        return self.json_impl.loads(MiniRacerValue(self, res).to_python())

    def heap_snapshot(self):
        """Return a snapshot of the V8 isolate heap."""

        with self.lock:
            res = self.ext.mr_heap_snapshot(self.ctx)

        return MiniRacerValue(self, res).to_python()

    def _free(self, res):
        self.ext.mr_free_value(self.ctx, res)

    def __del__(self):
        self.ext.mr_free_context(getattr(self, "ctx", None))


# Compatibility with versions 0.4 & 0.5
StrictMiniRacer = MiniRacer


class MiniRacerTypes(object):
    """MiniRacer types identifier

    Note: it needs to be coherent with mini_racer_extension.cc.
    """

    invalid = 0
    null = 1
    bool = 2
    integer = 3
    double = 4
    str_utf8 = 5
    array = 6  # deprecated
    hash = 7  # deprecated
    date = 8
    symbol = 9
    object = 10

    function = 100
    shared_array_buffer = 101
    array_buffer = 102

    execute_exception = 200
    parse_exception = 201
    oom_exception = 202
    timeout_exception = 203


class MiniRacerValueStruct(ctypes.Structure):
    _fields_ = [
        ("value", ctypes.c_void_p),  # value is 8 bytes, works only for 64bit systems
        ("type", ctypes.c_int),
        ("len", ctypes.c_size_t),
    ]


class ArrayBufferByte(ctypes.Structure):
    # Cannot use c_ubyte directly because it uses <B
    # as an internal type but we need B for memoryview.
    _fields_ = [("b", ctypes.c_ubyte)]
    _pack_ = 1


class MiniRacerValue(object):
    def __init__(self, ctx, ptr):
        self.ctx = ctx
        self.ptr = ptr

    def __str__(self):
        return str(self.to_python())

    @property
    def type(self):
        return self.ptr.contents.type

    @property
    def value(self):
        return self.ptr.contents.value

    @property
    def len(self):
        return self.ptr.contents.len

    def _double_value(self):
        ptr = ctypes.c_char_p.from_buffer(self.ptr.contents)
        return ctypes.c_double.from_buffer(ptr).value

    def _raise_from_error(self):
        if self.type == MiniRacerTypes.parse_exception:
            msg = ctypes.c_char_p(self.value).value
            raise JSParseException(msg)
        elif self.type == MiniRacerTypes.execute_exception:
            print(f"self.value:       {self.value}")
            msg = ctypes.c_char_p(self.value).value
            raise JSEvalException(msg.decode("utf-8", errors="replace"))
        elif self.type == MiniRacerTypes.oom_exception:
            msg = ctypes.c_char_p(self.value).value
            raise JSOOMException(msg)
        elif self.type == MiniRacerTypes.timeout_exception:
            msg = ctypes.c_char_p(self.value).value
            raise JSTimeoutException(msg)

    def to_python(self):
        self._raise_from_error()
        result = None
        typ = self.type
        if typ == MiniRacerTypes.null:
            result = None
        elif typ == MiniRacerTypes.bool:
            result = self.value == 1
        elif typ == MiniRacerTypes.integer:
            val = self.value
            result = 0 if val is None else ctypes.c_int32(val).value
        elif typ == MiniRacerTypes.double:
            result = self._double_value()
        elif typ == MiniRacerTypes.str_utf8:
            buf = ctypes.c_char_p(self.value)
            ptr = ctypes.cast(buf, ctypes.POINTER(ctypes.c_char))
            result = ptr[0 : self.len].decode("utf8")
        elif typ == MiniRacerTypes.function:
            result = JSFunction()
        elif typ == MiniRacerTypes.date:
            timestamp = self._double_value()
            # JS timestamp are milliseconds, in python we are in seconds
            result = datetime.datetime.utcfromtimestamp(timestamp / 1000.0)
        elif typ == MiniRacerTypes.symbol:
            result = JSSymbol()
        elif typ == MiniRacerTypes.shared_array_buffer or typ == MiniRacerTypes.array_buffer:
            cdata = (ArrayBufferByte * self.len).from_address(self.value)
            # Keep a reference to prevent the GC to free the backing store
            cdata._origin = self
            result = memoryview(cdata)
        elif typ == MiniRacerTypes.object:
            return JSObject(self.value)
        else:
            raise JSConversionException
        return result

    def __del__(self):
        self.ctx._free(self.ptr)



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  system.py
@Time    :  2022/5/2 2:52 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import contextlib
import csv
import json
import logging
import os
import platform
import shutil
import subprocess
import zipfile

import yaml

from custard.core.processor import DataKitHelper

with contextlib.suppress(AttributeError):
    # PyYAML version >= 5.1
    yaml.warnings({"YAMLLoadWarning": False})


logger = logging.getLogger(__name__)


class SystemHand(DataKitHelper):
    @classmethod
    def shell(cls, command):
        output, errors = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()
        o = output.decode("utf-8")
        return o

    @classmethod
    def get_depend_libs(cls, file_path):
        """
        获取第三方依赖架包
        Args:
            file_path:
        Returns:
        Examples:
        """
        if cls.is_fdir(file_path):
            with open(file_path, "r", encoding="utf-8") as file:
                content = file.readlines()
                format_con = ""
                for con in set(content):
                    if "#" not in con:
                        format_con += "".join("`{}`\t".format(con.strip()))
            return format_con
        return None

    @classmethod
    def get_platform_info(cls):
        """
        获取服务运行系统的基础信息
        Returns:
        Examples:
            >>> print(SystemHand.get_platform_info())
        """
        return {
            "platform": platform.platform(),  # 获取操作系统名称及版本号
            "machine": platform.machine(),  # 计算机类型
            "node": platform.node(),  # 计算机的网络名称
            "processor": platform.processor(),  # 计算机处理器信息
            "architecture": platform.architecture(),  # 获取操作系统的位数
            "python_version": platform.python_version(),  # 获取操作系统版本号
        }

    @classmethod
    def get_pool_config(cls, work_spaces_path, target_key=None, environment="pro"):
        """
        获取驱动配置
        Args:
            work_spaces_path:
            environment:
            target_key:
        Returns:
        Examples:
            >>> print(SystemHand.get_pool_config(target_key="email"))
        """
        application_file = "application.yaml" if environment == "pro" else f"application-{environment}.yaml"
        application_file_path = os.path.join(work_spaces_path, application_file)
        if os.path.exists(application_file_path):
            temp_data = SystemHand.load_file(application_file_path)
            result = temp_data[target_key] if target_key else temp_data
            return result, application_file_path
        else:
            raise FileNotFoundError(f"未找到application配置文件,请检查以下路径是否正确:{application_file_path}")

    @classmethod
    def is_fdir(cls, fdir_path, safe_loads=False):
        """
        文件是否存在
        Args:
            fdir_path:
            safe_loads: False 安全模式下可自动创建文件: True
        Returns:
        Examples:
            >>> print(SystemHand.is_fdir(fdir_path="Not Fund"))
            >>> print(SystemHand.is_fdir(fdir_path="../inutility"))
            >>> print(SystemHand.is_fdir(fdir_path="../inutility", safe_loads=True))
        """
        isfile, isdir = os.path.isfile(fdir_path), os.path.isdir(fdir_path)
        if isfile or isdir:
            return True
        elif safe_loads is False:
            return False
        else:
            with open(fdir_path, "w", encoding="utf-8") as file:
                file.close()
            return SystemHand.is_fdir(fdir_path=fdir_path)

    @classmethod
    def load_file(cls, file_path=None, load_mode="yaml"):
        """
        加载文件
        Args:
            file_path: 文件路径
            load_mode: 加载方式
        Returns:
        """
        with open(file_path, "r", encoding="utf-8") as file:
            if load_mode == "yaml":
                return yaml.safe_load(file)
            elif load_mode == "json":
                try:
                    return json.load(file)
                except json.JSONDecodeError as ex:
                    raise TypeError("JSONDecodeError:\nfile: %s\nerror: %s" % (file_path, ex))
            elif load_mode in (".txt", ".py"):
                return file
            elif load_mode == "html":
                return DataKitHelper.format_html_string(file)
            elif load_mode == "csv":
                csv_content_list = []
                with open(file_path, encoding="utf-8") as csvfile:
                    reader = csv.DictReader(csvfile)
                    for row in reader:
                        csv_content_list.append(row)
                return csv_content_list
            return None

    @classmethod
    def walk(cls, root_dir: str):
        """
        检索根目录,得到所有的子文件夹或子文件名
        Args:
            root_dir (str): _description_
        Examples:
            >>> root_dir = "../../custard"
            >>> SystemHand.walk(root_dir)
        """
        result = []
        dirs = [os.path.join(root_dir, index) for index in os.listdir(root_dir)]
        for index in dirs:
            for root, dirs, files in os.walk(index, topdown=False):
                file_path = [os.path.join(root, file) for file in files]
                result += file_path
        return result

    @classmethod
    def get_current_path(cls):
        """
        获取当前文件路径
        Returns:
        Examples:
            >>> SystemHand.get_current_path()
        """
        return os.path.abspath(os.path.dirname(__file__))

    @classmethod
    def get_superior_dir(cls):
        """
        获取上级目录
        Returns:
        Examples:
            >>> SystemHand.get_superior_dir()
        """
        return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

    @classmethod
    def get_dir_list(cls, file_path):
        """
        获取指定目录下所有的文件名并返回一个列表
        Args:
          file_path: 文件路径
        Returns:
        Examples:
            >>> SystemHand.get_dir_list()
        """
        current_files = os.listdir(file_path)
        all_files = []
        for file_name in current_files:
            full_file_name = os.path.join(file_path, file_name)
            all_files.append(full_file_name)
            if os.path.isdir(full_file_name):
                next_level_files = cls.get_dir_list(full_file_name)
                all_files.extend(next_level_files)
        return all_files

    @classmethod
    def tarfile(cls, file_path, method, target_path="./"):
        """
        打包文件为压缩包
        Args:
          file_path  被打包的文件路径
          target_path 目标存储的文件路径
          method    用于判断是打包单个文件还是遍历文件夹下所有的文件 single: 单个、 allie: 全部
        Returns:
        Examples:
        """
        ziplist = []
        zip_file = zipfile.ZipFile(target_path, "w")
        try:
            if not os.path.exists(file_path):
                raise FileNotFoundError("请检查file_path是否正确!")
            if method == "single":
                zip_file.write(file_path)
            if method == "allie":
                has_dir = not file_path.endswith(os.sep)
                if not has_dir:
                    raise FileNotFoundError("请检查file_path是否正确!")
                file_path = os.path.dirname(file_path)
                target_path = os.path.dirname(file_path) + os.sep + target_path
                if not os.path.exists(os.path.dirname(target_path)):
                    os.makedirs(os.path.dirname(target_path))
                # 多级目录读取
                for dirpath, _dirnames, filenames in os.walk(file_path):
                    for filename in filenames:
                        ziplist.append(os.path.join(dirpath, filename))
                for tar in ziplist:
                    zip_file.write(tar)
        except Exception as err:
            raise Exception(err)
        finally:
            zip_file.close()

    @classmethod
    def unzip(cls, method, file_path, target_path="./"):
        """
        解压多种类型的压缩包
        Args:
          method: 类型判断
          file_path: 需要解压的文件绝对路径
          zip_list: 获取压缩包内所有的文件
        Returns:

        Examples:
            >>> SystemHand.unzip(file_path="./tar.zip")
        """
        try:
            zip_file = zipfile.ZipFile(file_path)
            if method == "gzip":
                zip_file.extractall(path=target_path)
            elif method == "zip":
                zip_list = zip_file.namelist()
                for f in zip_list:
                    zip_file.extract(f, target_path)
            raise TypeError("Unknown method")
        except Exception as err:
            raise Exception(err)
        finally:
            zip_file.close()

    @classmethod
    def get_file_state(cls, file_path):
        """
        判断传入的文件状态
        Args:
          file_path: 需要解压的文件绝对路径
        Returns:
        Examples:
        """
        try:
            if os.access(file_path, os.F_OK):
                logger.info("%s:文件存在" % (file_path))
                if os.access(file_path, os.R_OK):
                    logger.info("%s:文件可读" % (file_path))
                else:
                    logger.error("%s:文件不支持可读" % (file_path))

                if os.access(file_path, os.W_OK):
                    logger.info("%s:文件可写" % (file_path))
                else:
                    logger.error("%s:文件不支持可写" % (file_path))

                if os.access(file_path, os.X_OK):
                    logger.info("%s:文件可执行" % (file_path))
                else:
                    logger.error("%s:文件不支持可执行" % (file_path))

                if os.path.isdir(file_path):
                    logger.info("%s:这是一个文件夹" % (file_path))
                else:
                    logger.info("%s:这是一个文件" % (file_path))
            else:
                logger.error("%s:文件不存在" % (file_path))
        except Exception as err:
            logger.error(err)
            raise Exception(err)

    @classmethod
    def copyfile(cls, file_path, target="./"):
        """
        复制文件
        Args:
          file_path: 需要解压的文件绝对路径
        Returns:
        Examples:
        """
        file_list = []
        if os.path.exists(file_path):
            if not os.path.exists(target):
                os.makedirs(target)
            # 多级目录读取
            for dirpath, _dirnames, filenames in os.walk(file_path):
                for filename in filenames:
                    file_list.append(os.path.join(dirpath, filename))
            for list in file_list:
                shutil.copy(list, target)
        else:
            logger.error("请检查file_path是否正确!")
        return target

    @classmethod
    def removefile(cls, file_path):
        """
        删除文件或文件夹
        Args:
          file_path: 需要解压的文件绝对路径
        Returns:
        Examples:
        """
        try:
            if os.path.exists(file_path):
                for root, dirs, files in os.walk(file_path, topdown=False):
                    # 先删除文件
                    for name in files:
                        os.remove(os.path.join(root, name))
                    # 再删除空目录
                    for name in dirs:
                        os.rmdir(os.path.join(root, name))
                        # 再删除自己
                        os.rmdir(file_path)
            else:
                logger.error("请检查file_path是否正确!")
        except Exception as FileNotFoundError:
            logger.error(FileNotFoundError)

    @classmethod
    def makefile(cls, file_path):
        """
        递归创建多级文件夹
        Args:
          file_path: 需要解压的文件绝对路径
        Returns:
        Examples:
        """
        try:
            if not os.path.exists(file_path):
                os.makedirs(file_path, exist_ok=True)
                if os.path.exists(file_path):
                    logger.info("目录:%s 创建成功!!!" % (file_path))
            else:
                logger.error("%s已存在,跳过创建!" % (file_path))
        except Exception as FileNotFoundError:
            logger.error(FileNotFoundError)

    @classmethod
    def depth_scan_file(cls, catalog, file_type):
        """
        过滤xxx目录下所有的xx格式文件
        Args:
          catalog: 指定目录
          file_type 类型
        Returns:
        Examples:
        """
        file_list = os.listdir(catalog)
        yaml_files = []
        for index in range(len(file_list)):
            name, suffix = os.path.splitext(file_list[index])
            # 判断文件类型
            if suffix.replace(".", "").lower() == file_type:
                yaml_files.append("%s\\%s" % (catalog, file_list[index]))
        return yaml_files

    @classmethod
    def get_file_type(cls, file_path):
        """
        获取文件类型
        Args:
          file_path: 指定目录
        Returns:
        Examples:
        """
        if file_path is None or os.path.isfile(file_path) is False:
            raise FileNotFoundError("Please check whether the file path or file name exists")
        head, tail = os.path.split(file_path)
        name, suffix = os.path.splitext(tail)
        return suffix.lower()

    @classmethod
    def json_to_yaml(cls, json_file):
        """
        json文件格式转yaml
        """
        if json_file.endswith("json"):
            with open(json_file, "r") as pf:
                json_to_dict = cls.safely_json_loads(pf.read())
            yaml_file = json_file.replace(".json", ".yaml")
            with open(yaml_file, "w") as fp:
                yaml.safe_dump(json_to_dict, stream=fp, default_flow_style=False)
                return True
        return False

    @classmethod
    def yaml_to_json(cls, yaml_file):
        """
        yaml文件格式转json
        """
        if yaml_file.endswith("yaml"):
            with open(yaml_file, "r") as pf:
                load_data_ = yaml.load(pf, Loader=yaml.FullLoader)
                json_data_ = json.dumps(
                    load_data_,
                    sort_keys=False,
                    ensure_ascii=False,
                    indent=4,
                    separators=(",", ": "),
                )
            json_file = yaml_file.replace(".yaml", ".json")
            with open(json_file, "w") as fp:
                fp.write(json_data_)
                return True
        else:
            return False

    @classmethod
    def write_json(cls, data, json_path, method="w"):
        """
        把处理后的参数写入json文件
        :param res:
        :param json_path:
        :return:
        """
        if isinstance(data, (dict, list)):
            with open(json_path, method, encoding="utf-8") as f:
                json.dump(data, f, ensure_ascii=False, sort_keys=True, indent=4)
                return True
        return False



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from .dafunc import TimerContextManager
from .dafunc import bind_timeout
from .dafunc import calc_time
from .dafunc import func_timeout
from .exceptions import FunctionTimedOut
from .moment import Moment
from .stoppable_thread import StoppableThread

__all__ = [
    "TimerContextManager",
    "StoppableThread",
    "bind_timeout",
    "calc_time",
    "func_timeout",
    "FunctionTimedOut",
    "Moment",
]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  area_code.py
@Time    :  2022/5/30 1:35 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  身份证效验&随机生成身份证信息
"""

import random
import re
from datetime import datetime, timedelta

from custard.time.moment import Moment

from .area_code import AREA_INFO, ID_NUMBER_15_REGEX, ID_NUMBER_18_REGEX


class IdNumber(str):
    def __init__(self, id_number=None):
        super(IdNumber, self).__init__()
        self.id = id_number
        self.area_id = int(self.id[0:6])
        self.birth_year = int(self.id[6:10])
        self.birth_month = int(self.id[10:12])
        self.birth_day = int(self.id[12:14])

    def get_area_name(self):
        """根据区域编号取出区域名称"""
        return AREA_INFO[self.area_id]

    def get_birthday(self):
        """通过身份证号获取出生日期"""
        return "{0}-{1}-{2}".format(self.birth_year, self.birth_month, self.birth_day)

    def get_age(self):
        """通过身份证号获取年龄"""
        now = datetime.now() + timedelta(days=1)
        year, month, day = now.year, now.month, now.day

        if year == self.birth_year:
            return 0
        else:
            if self.birth_month > month or (self.birth_month == month and self.birth_day > day):
                return year - self.birth_year - 1
            else:
                return year - self.birth_year

    def get_sex(self):
        """通过身份证号获取性别, 女生:0,男生:1"""
        return int(self.id[16:17]) % 2

    def get_check_digit(self):
        """通过身份证号获取校验码"""
        check_sum = 0
        for i in range(0, 17):
            check_sum += ((1 << (17 - i)) % 11) * int(self.id[i])
        check_digit = (12 - (check_sum % 11)) % 11
        return check_digit if check_digit < 10 else "X"

    @classmethod
    def check_verify_id(cls, id_number) -> bool:
        """校验身份证是否正确"""
        if re.match(ID_NUMBER_18_REGEX, id_number):
            check_digit = cls(id_number).get_check_digit()
            return str(check_digit) == id_number[-1]
        else:
            return bool(re.match(ID_NUMBER_15_REGEX, id_number))

    @classmethod
    def get_id_card(cls, sex=0, end_date=Moment.get_now_time("%Y-%m-%d")):
        """随机生成身份证号,sex = 0表示女性,sex = 1表示男性"""
        # 随机生成一个区域码(6位数)
        id_number = str(random.choice(list(AREA_INFO.keys())))
        # 限定出生日期范围(8位数)
        start = datetime.strptime("1960-01-01", "%Y-%m-%d")
        end = datetime.strptime(end_date, "%Y-%m-%d")
        birth_days = datetime.strftime(start + timedelta(random.randint(0, (end - start).days + 1)), "%Y%m%d")
        id_number += str(birth_days)
        # 顺序码(2位数)
        id_number += str(random.randint(10, 99))
        # 性别码(1位数)
        id_number += str(random.randrange(sex, 10, step=2))
        # 校验码(1位数)
        return id_number + str(cls(id_number).get_check_digit())


if __name__ == "__main__":
    random_sex = random.randint(0, 1)  # 随机生成男(1)或女(0)
    id_card = IdNumber.get_id_card(random_sex)  # 随机生成身份证号
    print(IdNumber(id_card).area_id)  # 地址编码:431121
    print(IdNumber(id_card).get_area_name())  # 地址:湖南省永州市祁阳县
    print(IdNumber(id_card).get_birthday())  # 生日:1999-3-15
    print(IdNumber(id_card).get_age())  # 年龄:21(岁)
    print(IdNumber(id_card).get_sex())  # 性别:1(男)
    print(IdNumber(id_card).get_check_digit())  # 校验码:1
    print(IdNumber.check_verify_id("411424197912081942"))



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import aioredis
import uvicorn
from fastapi import Depends, FastAPI, HTTPException, WebSocket

from custard.limiter import Limiter
from custard.limiter.depends import RateLimiter, WebSocketRateLimiter

app = FastAPI()


@app.on_event("startup")
async def startup():
    r = aioredis.from_url("redis://localhost", password="M5Pi9YW6u", db=0, port=16389)
    await Limiter.init(r)


@app.on_event("shutdown")
async def shutdown():
    await Limiter.close()


@app.get("/", dependencies=[Depends(RateLimiter(times=2, seconds=5))])
async def index_get():
    return {"msg": "Hello World"}


@app.post("/", dependencies=[Depends(RateLimiter(times=1, seconds=5))])
async def index_post():
    return {"msg": "Hello World"}


@app.get(
    "/multiple",
    dependencies=[
        Depends(RateLimiter(times=1, seconds=5)),
        Depends(RateLimiter(times=2, seconds=15)),
    ],
)
async def multiple():
    return {"msg": "Hello World"}


@app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    ratelimit = WebSocketRateLimiter(times=1, seconds=5)
    while True:
        try:
            data = await websocket.receive_text()
            await ratelimit(websocket, context_key=data)  # NB: context_key is optional
            await websocket.send_text("Hello, world")
        except HTTPException:
            await websocket.send_text("Hello again")


if __name__ == "__main__":
    uvicorn.run("__init__:app", reload=True)



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2020/9/25 19:55
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from custard.curl.parse import curl_to_request_kwargs

__all__ = ["curl_to_request_kwargs"]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  This library is provided to allow standard python logging to output log data as JSON formatted strings
"""
from .captcha import Captcha, CaptchaPainter

__all__ = ["Captcha", "CaptchaPainter"]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  streambody.py
@Time    :  2020/9/25 19:55
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""

import os
import uuid


class StreamBody(object):
    def __init__(self, rt):
        self._rt = rt
        self._read_len = 0
        self._content_len = 0
        self._use_chunked = False
        self._use_encoding = False
        if "Content-Length" in self._rt.headers:
            self._content_len = int(self._rt.headers["Content-Length"])
        elif "Transfer-Encoding" in self._rt.headers and self._rt.headers["Transfer-Encoding"] == "chunked":
            self._use_chunked = True
        else:
            raise IOError("create StreamBody failed without Content-Length header or Transfer-Encoding header")

        if "Content-Encoding" in self._rt.headers:
            self._use_encoding = True

    def __iter__(self):
        """提供一个默认的迭代器"""
        return self._rt.iter_content(1024)

    def __len__(self):
        return self._content_len

    def get_raw_stream(self):
        """提供原始流"""
        return self._rt.raw

    def get_stream(self, chunk_size=1024):
        """提供一个chunk可变的迭代器"""
        return self._rt.iter_content(chunk_size=chunk_size)

    def read(self, chunk_size=1024, auto_decompress=False):
        chunk = None
        if self._use_encoding and not auto_decompress:
            chunk = self._rt.raw.read(chunk_size)
        else:
            try:
                chunk = next(self._rt.iter_content(chunk_size))
            except StopIteration:
                return ""
        return chunk

    def get_stream_to_file(self, file_name, auto_decompress=False):
        """保存流到本地文件"""
        self._read_len = 0
        tmp_file_name = "{file_name}_{uuid}".format(file_name=file_name, uuid=uuid.uuid4().hex)
        with open(tmp_file_name, "wb") as fp:
            while 1:
                chunk = self.read(1024, auto_decompress)
                if not chunk:
                    break
                self._read_len += len(chunk)
                fp.write(chunk)

        if (
            not self._use_chunked
            and not (self._use_encoding and auto_decompress)
            and self._read_len != self._content_len
        ):
            if os.path.exists(tmp_file_name):
                os.remove(tmp_file_name)
            raise IOError("download failed with incomplete file")
        if os.path.exists(file_name):
            os.remove(file_name)
        os.rename(tmp_file_name, file_name)

    def pget_stream_to_file(self, fdst, offset, expected_len, auto_decompress=False):
        """保存流到本地文件的offset偏移"""
        self._read_len = 0
        fdst.seek(offset, 0)
        chunk_size = 1024 * 1024
        while 1:
            chunk = self.read(chunk_size, auto_decompress)
            if not chunk:
                break
            self._read_len += len(chunk)
            fdst.write(chunk)

        if not self._use_chunked and not (self._use_encoding and auto_decompress) and self._read_len != expected_len:
            raise IOError("download failed with incomplete file")



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  iterables.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from itertools import islice
from typing import Generic, Iterable, Optional, TypeVar

from pydantic import conint

from .api import create_page, resolve_params
from .bases import AbstractPage, AbstractParams
from .limit_offset import LimitOffsetPage as DefaultLimitOffsetPage
from .limit_offset import LimitOffsetParams
from .pagination import Page as DefaultPage
from .pagination import Params

T = TypeVar("T")


class Page(DefaultPage, Generic[T]):
    total: Optional[conint(ge=0)]  # type: ignore


class LimitOffsetPage(DefaultLimitOffsetPage, Generic[T]):
    total: Optional[conint(ge=0)]  # type: ignore


def paginate(
    iterable: Iterable[T],
    params: Optional[AbstractParams] = None,
    total: Optional[int] = None,
) -> AbstractPage[T]:
    params = resolve_params(params)
    raw_params = params.to_raw_params()

    items = [*islice(iterable, raw_params.offset, raw_params.offset + raw_params.limit)]
    return create_page(items, total, params)  # type: ignore


__all__ = [
    "Page",
    "Params",
    "LimitOffsetPage",
    "LimitOffsetParams",
    "paginate",
]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  pagination.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from __future__ import annotations

import math
from typing import Generic, Sequence, TypeVar

from fastapi import Query
from pydantic import BaseModel

from .bases import AbstractPage, AbstractParams, RawParams

T = TypeVar("T")


class Params(BaseModel, AbstractParams):
    page: int = Query(1, ge=1, description="Page number")
    size: int = Query(20, gt=0, le=100, description="Page size")

    def to_raw_params(self) -> RawParams:
        return RawParams(
            limit=self.size,
            offset=self.size * (self.page - 1),
        )


class Page(AbstractPage[T], Generic[T]):
    results: Sequence[T]
    total: int
    page: int
    size: int
    next: str
    previous: str
    total_pages: int

    __params_type__ = Params  # Set params related to Page

    @classmethod
    def create(
        cls,
        results: results,
        total: int,
        params: Params,
    ) -> Page[T]:
        page = params.page
        size = params.size
        total_pages = math.ceil(total / params.size)
        next = f"?page={page + 1}&size={size}" if (page + 1) <= total_pages else "null"
        previous = f"?page={page - 1}&size={size}" if (page - 1) >= 1 else "null"

        return cls(
            results=results,
            total=total,
            page=params.page,
            size=params.size,
            next=next,
            previous=previous,
            total_pages=total_pages,
        )



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  test_depends.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from time import sleep

from starlette.testclient import TestClient

from custard.limiter.examples import app


def test_limiter() -> None:
    with TestClient(app) as client:
        response = client.get("/")
        assert response.status_code == 200

        response = client.get("/")
        assert response.status_code == 200

        response = client.get("/")
        assert response.status_code == 429

        response = client.post("/")
        assert response.status_code == 200

        response = client.post("/")
        assert response.status_code == 429
        sleep(5)

        response = client.get("/")
        assert response.status_code == 200

        response = client.post("/")
        assert response.status_code == 200


def test_limiter_multiple() -> None:
    with TestClient(app) as client:
        response = client.get("/multiple")
        assert response.status_code == 200

        response = client.get("/multiple")
        assert response.status_code == 429
        sleep(5)

        response = client.get("/multiple")
        assert response.status_code == 200

        response = client.get("/multiple")
        assert response.status_code == 429
        sleep(10)

        response = client.get("/multiple")
        assert response.status_code == 200


def test_limiter_websockets() -> None:
    with TestClient(app) as client, client.websocket_connect("/ws") as ws:
        ws.send_text("Hi")
        data = ws.receive_text()
        assert data == "Hello, world"

        ws.send_text("Hi")
        data = ws.receive_text()
        assert data == "Hello again"

        ws.send_text("Hi 2")
        data = ws.receive_text()
        assert data == "Hello, world"
        ws.close()



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  limit_offset.py
@Time    :  2022/5/1 8:21 PM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from __future__ import annotations

from dataclasses import asdict
from typing import Generic, Sequence, TypeVar

from fastapi import Query
from pydantic import BaseModel
from pydantic.types import conint

from .bases import AbstractParams, BasePage, RawParams

T = TypeVar("T")


class LimitOffsetParams(BaseModel, AbstractParams):
    limit: int = Query(50, ge=1, le=100, description="Page size limit")
    offset: int = Query(0, ge=0, description="Page offset")

    def to_raw_params(self) -> RawParams:
        return RawParams(
            limit=self.limit,
            offset=self.offset,
        )


class LimitOffsetPage(BasePage[T], Generic[T]):
    limit: conint(ge=1)  # type: ignore
    offset: conint(ge=0)  # type: ignore

    __params_type__ = LimitOffsetParams

    @classmethod
    def create(
        cls,
        items: Sequence[T],
        total: int,
        params: AbstractParams,
    ) -> LimitOffsetPage[T]:
        return cls(
            total=total,
            items=items,
            **asdict(params.to_raw_params()),
        )


Page = LimitOffsetPage
Params = LimitOffsetParams

__all__ = [
    "Page",
    "Params",
    "LimitOffsetPage",
    "LimitOffsetParams",
]



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/7/18 10:16
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
from requests import request
from custard.core.system import SystemHand
from .exception import ParseMethodError
from .swagger import Swagger2


def load_url(url, method="get", **kwargs):
    """
    通过url加载json
    Args:
        url:
        method:
        **kwargs:

    Returns:

    """
    return request(url=url, method=method, **kwargs).json()


def swagger_parse(url=None, file=None, deep=5, **kwargs):
    """
    解析swagger
    Args:
        url:
        file:
        deep:
        **kwargs:

    Returns:

    """
    if url:
        source = load_url(url, **kwargs)
    elif file:
        source = SystemHand.load_file(file, "json")
    else:
        raise ParseMethodError("解析方式错误")
    return Swagger2(source, deep=deep)



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  __init__.py
@Time    :  2022/5/27 12:09 AM
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  None
"""
import codecs
import os
import typing
from json import JSONEncoder

from ..mock import mini_racer

MiniRacer = mini_racer.MiniRacer


class Mock:
    def __init__(self):
        self.__code = codecs.open(os.path.join(os.path.dirname(__file__), "mock.min.js"), encoding="utf-8").read()
        self.__ctx = MiniRacer()
        self.__ctx.eval(self.__code)

    def mock(
        self,
        template: typing.Union[dict, list, str],
        encoder=JSONEncoder,
        timeout=0,
        max_memory=0,
    ) -> typing.Union[dict, list, str]:
        """
        Mock from python object
        :param template: Mock template
        :param encoder: You can pass a custom JSON encoder by passing it in the encoder
        :param timeout: Limit run timeout, default no limit: timeout = 0(millisecond)
        :param max_memory: Limit max memory, default no limit: max_memory = 0
        :return: dict, list, str
        """
        return self.__ctx.call("Mock.mock", template, encoder=encoder, timeout=timeout, max_memory=max_memory)

    def mock_js(self, js_str: str, timeout=0, max_memory=0) -> typing.Union[dict, list, str]:
        """
        Mock form JSON string or JavaScript Object like-string
        :param js_str: Mock template
        :param timeout: Limit run timeout, default no limit: timeout = 0(millisecond)
        :param max_memory: Limit max memory, default no limit: max_memory = 0
        :return: dict, list, str
        """
        js = "Mock.mock({template})".format(template=js_str)
        return self.__ctx.eval(js, timeout, max_memory)


mocker = Mock()



# -*- coding:utf-8 -*-
# !/usr/bin/env python 3.9.11
"""
@File    :  regex.py
@Time    :  2021/10/22 20:16
@Author  :  YuYanQing
@Version :  1.0
@Contact :  mryu168@163.com
@License :  (C)Copyright 2022-2026
@Desc    :  正则匹配
"""

import re
from typing import Any


class RegEx:
    @staticmethod
    def match_email(context: Any) -> bool:
        """
        效验邮箱格式
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["test@163.com","test163.com","155555@qq.com"]
            >>> BatchTask.list_jobs(RegEx.match_email, examples)
        """
        return bool(re.match("^.+\\@(\\[?)[a-zA-Z0-9\\-\\.]+\\.([a-zA-Z]{2,3}|[0-9]{1,3})(\\]?)", str(context)))

    @staticmethod
    def match_double_byte_str(context: Any) -> bool:
        """
        效验是否存在双字节
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈"]
            >>> BatchTask.list_jobs(RegEx.match_double_byte_str, examples)
        """
        return bool(re.match(".*?([^x00-xff])", str(context)))

    @staticmethod
    def weak_pwd(context: Any) -> bool:
        """
        效验密码是否为弱密码 (最低要求数字、英文、符合各一个、长度限制:7~20)
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None, "QSa12356@"]
            >>> BatchTask.list_jobs(RegEx.weak_pwd, examples)
        """
        return re.match("^(?:(?=.*[0-9].*)(?=.*[A-Za-z].*)(?=.*[\\W].*))[\\W0-9A-Za-z]{7,20}", str(context)) is None

    @staticmethod
    def match_mobile(context: Any) -> bool:
        """
        效验手机号格式
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None, "QSa12356@",
            ... "+8617888829981","008618311006933","19119255552"]
            >>> BatchTask.list_jobs(RegEx.match_mobile, examples)
        """
        rule = (
            "^(?:(?:\\+|00)86)?1(?:(?:3[\\d])|(?:4[5-79])|"
            "(?:5[0-35-9])|(?:6[5-7])|(?:7[0-8])|(?:8[\\d])|(?:9[189]))\\d{8}"
        )
        return bool(re.match(rule, str(context)))

    @staticmethod
    def match_ipv4(context: Any) -> bool:
        """
        效验ipv4格式
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None,
            ... "QSa12356@", "+8617888829981","008618311006933","19119255552", "127.16.0.0"]
            >>> BatchTask.list_jobs(RegEx.match_ipv4, examples)
        """
        rule = r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
        return bool(re.match(rule, str(context)))

    @staticmethod
    def match_str_length(context: Any, min_length=15, max_length=17) -> bool:
        """
        效验数字长度是否在区间值内
        Args:
            context:
            min_length:
            max_length:
        Returns:
        Examples:
            >>> examples = [17, 12355678901235567, "哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None,
            ... "QSa12356@", "+8617888829981","008618311006933","19119255552", "127.16.0.0"]
            >>> BatchTask.list_jobs(RegEx.match_str_length, examples)
        """
        rule = "^\\d{%s,%s}" % (min_length, max_length)
        return bool(re.match(rule, str(context)))

    @staticmethod
    def match_username(context, min_length=7, max_length=20):
        """
        效验用户名格式 (必须由英文开头、长度限制:7~20)
        Args:
            context:
            min_length:
            max_length:
        Returns:
        Examples:
            >>> examples = [17, 12355678901235567, "QS356", "哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None,
            ... "QSa12356@", "+8617888829981","008618311006933","19119255552", "127.16.0.0"]
            >>> BatchTask.list_jobs(RegEx.match_username, examples)
        """
        rule = r"^(?=.*[A-Za-z])[a-zA-Z0-9]{%s,%s}" % (min_length, max_length)
        return bool(re.match(rule, str(context)))

    @staticmethod
    def match_valid_url(context):
        """
        效验url
        context:
        Returns:
        Examples:
            >>> examples = ["https://www.baidu.com", "127.0.0.1:8000", "https://www.sweets.cn:8080", "127.16.0.0",
            ... 17, 12355678901235567, "QS356", "哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None,
            ... "QSa12356@", "+8617888829981","008618311006933","19119255552"]
            >>> BatchTask.list_jobs(RegEx.match_valid_url, examples)
        """
        rule = r"^(((ht|f)tps?):\/\/)?[\w-]+(\.[\w-]+)+([\w.@?^=%&:/~+#-]*[\w@?^=%&/~+#-])?$"
        return bool(re.match(rule, str(context)))

    @staticmethod
    def match_trail_type(context, method=None):
        """
        效验尾缀格式
        Args:
            context:
            method:
        Returns:
        Examples:
            >>> examples = [".bat", ".image", ".png", "image"]
            >>> BatchTask.list_jobs(RegEx.match_trail_type, examples)
        """
        image = ".*(\\.png|\\.jpg|\\.jpeg|\\.gif|\\.mov)$"
        video = ".*(\\.mp4|\\.avi|\\.mkv|\\.flv|\\.vob)$"
        exe = ".*(\\.exe|\\.sh|\\.bat)$"
        docs = ".*(\\.md|\\.xls|\\.xlsx|\\.word|\\.pdf)$"
        if method == "image":
            rule = image
        elif method == "video":
            rule = video
        elif method == "docs":
            rule = docs
        else:
            rule = exe
        return bool(re.match(rule, str(context).lower()))

    @staticmethod
    def match_train_number(context: Any) -> bool:
        """
         火车车次
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_train_number('G1868')
            >>> examples = ['G1868', 'D102', 'D9', 'Z5', 'Z26', 'Z17']
            >>> BatchTask.list_jobs(RegEx.match_train_number, examples)
        """
        return bool(re.match("^[GCDZTSPKXLY1-9]\\d{1,4}$", str(context)))

    @staticmethod
    def match_phone_imei(context: Any) -> bool:
        """
         手机机身码(IMEI)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_phone_imei('1235567890')
            >>> RegEx.match_phone_imei('123556789012355')
            >>> examples = ['123556789012355', '1235567890123556', '12355678901235567']
            >>> BatchTask.list_jobs(RegEx.match_phone_imei, examples)
        """
        return bool(re.match("^\\d{15,17}$", str(context)))

    @staticmethod
    def match_ip(context: Any) -> bool:
        """
         必须带端口号的网址(或ip)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_ip('127.0.0.1:5050')
            >>> RegEx.match_ip('https://www.123.com')
            >>> RegEx.match_ip('https://www.123.com:8080')
            >>> examples = ['https://www.qq.com:8080', '127.0.0.1:5050', 'baidu.com:8001', 'http://192.168.1.1:9090']
            >>> BatchTask.list_jobs(RegEx.match_ip, examples)
        """
        return bool(re.match("^((ht|f)tps?:\\/\\/)?[\\w-]+(\\.[\\w-]+)+:\\d{1,5}\\/?$", str(context)))

    @staticmethod
    def match_url(context: Any) -> bool:
        """
         网址(URL)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_url('www.123.com')
            >>> RegEx.match_url('https://vuejs.org/v2/api/#v-model')
            >>> RegEx.match_url('//www.123.com')
            >>> examples = ['ftp://baidu.123', 'https://www.amap.com/search?id=BV10060895&city=420111']
            >>> BatchTask.list_jobs(RegEx.match_url, examples)
        """
        return bool(
            re.match(
                "^(((ht|f)tps?):\\/\\/)?([^!@#$%^&*?.\\s-]([^!@#$%^&*?.\\s]{0,63}[^!@#$%^&*?.\\s])?\\.)+[a-z]{2,6}\\/?",
                str(context),
            ),
        )

    @staticmethod
    def match_social_credit(context: Any) -> bool:
        """
         统一社会信用代码
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_social_credit('9111010')
            >>> RegEx.match_social_credit('92371000MA3MXH0E3W')
            >>> examples = ['91230186MA1B7FLT55', '92371000MA3MXH0E3W']
            >>> BatchTask.list_jobs(RegEx.match_social_credit, examples)
        """
        return bool(re.match("^[0-9A-HJ-NPQRTUWXY]{2}\\d{6}[0-9A-HJ-NPQRTUWXY]{10}$", str(context)))

    @staticmethod
    def match_easy_social_credit(context: Any) -> bool:
        """
         统一社会信用代码(宽松匹配)(15位/18位/20位数字/字母)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_easy_social_credit('91110108772551611J')
            >>> examples = ['91110108772551611J', '911101085923662400']
            >>> BatchTask.list_jobs(RegEx.match_easy_social_credit, examples)
        """
        return bool(re.match("^(([0-9A-Za-z]{15})|([0-9A-Za-z]{18})|([0-9A-Za-z]{20}))$", str(context)))

    @staticmethod
    def match_net_mask(context: Any) -> bool:
        """
         子网掩码(不包含 0.0.0.0)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_net_mask('2')
            >>> RegEx.match_net_mask('255.255.255.0')
            >>> examples = ['255.255.255.0', '255.255.255.255', '255.240.0.0']
            >>> BatchTask.list_jobs(RegEx.match_net_mask, examples)
        """
        return bool(
            re.match(
                "^(254|252|248|240|224|192|128)\\.0\\.0\\.0|255\\.(254|252|248|240|224|192|128|0)\\.0\\.0|255\\.255\\.(254|252|248|240|224|192|128|0)\\.0|255\\.255\\.255\\.(255|254|252|248|240|224|192|128|0)$",
                str(context),
            ),
        )

    @staticmethod
    def match_md5_format(context: Any) -> bool:
        """
         md5格式(32位)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_md5_format('21fe181c5bfc16306a6828c1f7b762e8')
            >>> examples = ['21fe181c5bfc','21fe181c5bfc16306a6828c1f7b762e8']
            >>> BatchTask.list_jobs(RegEx.match_md5_format, examples)
        """
        return bool(re.match("^([a-f\\d]{32}|[A-F\\d]{32})$", str(context)))

    @staticmethod
    def match_uuid_format(context: Any) -> bool:
        """
         GUID/UUID
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_uuid_format('51E3DAF5-6E37-4BCC-9F8E-0D9521E2AA8D')
            >>> examples = ['21fe181c5bfc','21fe181c5bfc16306a6828c1f7b762e8','51E3DAF5-6E37-4BCC-9F8E-0D9521E2AA8D']
            >>> BatchTask.list_jobs(RegEx.match_uuid_format, examples)
        """
        return bool(re.match("^[a-f\\d]{4}(?:[a-f\\d]{4}-){4}[a-f\\d]{12}$/i", str(context)))

    @staticmethod
    def match_version_format(context: Any) -> bool:
        """
         版本号(version)格式必须为X.Y.Z
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_version_format('16.5')
            >>> RegEx.match_version_format('16.5.16')
        """
        return bool(re.match("^\\d+(?:\\.\\d+){2}$", str(context)))

    @staticmethod
    def match_image_url_format(context: Any) -> bool:
        """
         图片(image)链接地址(图片格式可按需增删)
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['www.123.com/logo.png', 'https://www.abc.com/logo.png']
            >>> BatchTask.list_jobs(RegEx.match_image_url_format, examples)
        """
        return bool(re.match("^https?:\\/\\/(.+\\/)+.+(\\.(gif|png|jpg|jpeg|webp|svg|psd|bmp|tif))$", str(context)))

    @staticmethod
    def match_video_url_format(context: Any) -> bool:
        """
         视频(video)链接地址(视频格式可按需增删)
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['www.123.com/lwc.avi', 'https://www.abc.com/logo.mpg']
            >>> BatchTask.list_jobs(RegEx.match_video_url_format, examples)
        """
        return bool(
            re.match("^https?:\\/\\/(.+\\/)+.+(\\.(swf|avi|flv|mpg|rm|mov|wav|asf|3gp|mkv|rmvb|mp4))$", str(context)),
        )

    @staticmethod
    def match_24hms_time_format(context: Any) -> bool:
        """
         24小时制时间(HH:mm:ss)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_24hms_time_format('23:35:55')
        """
        return bool(re.match("^(?:[01]\\d|2[0-3]):[0-5]\\d:[0-5]\\d$", str(context)))

    @staticmethod
    def match_12hms_time_format(context: Any) -> bool:
        """
         12小时制时间(HH:mm:ss)
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_12hms_time_format('01:00:00')
            >>> RegEx.match_12hms_time_format('12:00:00')
            >>> RegEx.match_12hms_time_format('23:35:55')
        """
        return bool(re.match("^(?:1[0-2]|0?[1-9]):[0-5]\\d:[0-5]\\d$", str(context)))

    @staticmethod
    def match_base64_format(context: Any) -> bool:
        """
         base64格式
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_base64_format('data:image/gif;base64,xxxx==')
        """
        return bool(
            re.match(
                "^\\s*data:(?:[a-z]+\\/[a-z0-9-+.]+(?:;[a-z-]+=[a-z0-9-]+)?)?(?:;base64)?,([a-z0-9!$&',()*+;=\\-._~:@/?%\\s]*?)\\s*$",
                str(context),
            ),
        )

    @staticmethod
    def match_easy_currency_format(context: Any) -> bool:
        """
         数字/货币金额(支持负数、千分位分隔符)
        Args:
            context:
        Returns:
        Examples:
            >>> examples = [100, -0.99, 3, 234.32, -1, 900, 235.09, '12,345,678.90']
            >>> BatchTask.list_jobs(RegEx.match_easy_currency_format, examples)
        """
        return bool(re.match("^-?\\d+(,\\d{3})*(\\.\\d{1,2})?$", str(context)))

    @staticmethod
    def match_chinese_name(context: Any) -> bool:
        """
         中文姓名
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['葛二蛋', '凯文·杜兰特', '德克·维尔纳·诺维茨基']
            >>> BatchTask.list_jobs(RegEx.match_chinese_name, examples)
        """
        return bool(re.match("^(?:[一-龥·]{2,16})$", str(context)))

    @staticmethod
    def match_english_name(context: Any) -> bool:
        """
         英文姓名
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['James', 'Kevin Wayne Durant', 'Dirk Nowitzki']
            >>> BatchTask.list_jobs(RegEx.match_english_name, examples)
        """
        return bool(re.match("(^[a-zA-Z][a-zA-Z\\s]{0,20}[a-zA-Z]$)", str(context)))

    @staticmethod
    def match_only_chinese(context: Any) -> bool:
        """
         只有中文/汉字
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["QS12356","哈哈哈","155555@qq.com","english"]
            >>> BatchTask.list_jobs(RegEx.match_only_chinese, examples)
        """
        return bool(
            re.match(
                "^(?:[㐀-䶵一-鿪﨎﨏﨑﨓﨔﨟﨡﨣﨤﨧-﨩]|[�-��-��-��-�][�-�]|�[�-��-�]|�[�-��-�]|�[�-��-�]|�[�-��-�]|�[�-�])+$",
                str(context),
            ),
        )

    @staticmethod
    def match_only_english(context: Any) -> bool:
        """
         英文字母
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["https://www.baidu.com", "127.0.0.1:8000", "https://www.sweets.cn:8080", "127.16.0.0",
            ... 17, 12355678901235567, "QS356", "哈哈哈", "12356", "abc", "ABC" "QS12356", "QS12356哈哈哈", None,
            ... "QSa12356@", "+8617888829981","008618311006933","19119255552"]
            >>> BatchTask.list_jobs(RegEx.match_only_english, examples)
        """
        return bool(re.match("^[a-zA-Z]+$", str(context)))

    @staticmethod
    def match_only_decimals(context: Any) -> bool:
        """
         小数
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['0.0', '0.09']
            >>> BatchTask.list_jobs(RegEx.match_only_decimals, examples)
        """
        return bool(re.match("^\\d+\\.\\d+$/", str(context)))

    @staticmethod
    def match_only_number(context: Any) -> bool:
        """
         只包含数字
        Args:
            context:
        Returns:
        Examples:
            >>> RegEx.match_only_number(1235678)
            >>> RegEx.match_only_number('1235678')
            >>> RegEx.match_only_number('class')
            >>> RegEx.match_only_number('啊啊呸啊呸')
        """
        return bool(re.match("^\\d+$", str(context)))

    @staticmethod
    def match_only_number_and_letters(context: Any) -> bool:
        """
         数字和字母组成
        Args:
            context:
        Returns:
        Examples:
            >>> examples =  ['james666', '125666', '啊呸']
            >>> BatchTask.list_jobs(RegEx.match_only_number_and_letters, examples)
        """
        return bool(re.match("^[A-Za-z0-9]+$", str(context)))

    @staticmethod
    def match_only_lowe_letters(context: Any) -> bool:
        """
         小写英文字母组成
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['russel', 'ABC', '1235678', 'Ab123']
            >>> BatchTask.list_jobs(RegEx.match_only_lowe_letters, examples)
        """
        return bool(re.match("^[a-z]+$", str(context)))

    @staticmethod
    def match_only_upper_letters(context: Any) -> bool:
        """
         大写英文字母
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['russel', 'ABC', '1235678', 'Ab123']
            >>> BatchTask.list_jobs(RegEx.match_only_upper_letters, examples)
        """
        return bool(re.match("^[A-Z]+$", str(context)))

    @staticmethod
    def match_only_chinese_and_number(context: Any) -> bool:
        """
         中文和数字
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['哈哈哈', '你好6777啊', 'abc', 'ABC', '1235678', '@¥()!']
            >>> BatchTask.list_jobs(RegEx.match_only_chinese_and_number, examples)
        """
        return bool(
            re.match(
                "^((?:[㐀-䶵一-鿪﨎﨏﨑﨓﨔﨟﨡﨣﨤﨧-﨩]|[�-��-��-��-�][�-�]|�[�-��-�]|�[�-��-�]|�[�-��-�]|�[�-��-�]|�[�-�])|(\\d))+$",
                str(context),
            ),
        )

    @staticmethod
    def match_not_exists_letters(context: Any) -> bool:
        """
         不能包含字母
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ['哈哈哈', '你好6啊', 'abc', 'ABC', '1235678', '@¥()!']
            >>> BatchTask.list_jobs(RegEx.match_not_exists_letters, examples)
        """
        return bool(
            re.match(
                "^((?:[㐀-䶵一-鿪﨎﨏﨑﨓﨔﨟﨡﨣﨤﨧-﨩]|[�-��-��-��-�][�-�]|�[�-��-�]|�[�-��-�]|�[�-��-�]|�[�-��-�]|�[�-�])|(\\d))+$",
                str(context),
            ),
        )

    @staticmethod
    def match_ascii_special_char(context: Any) -> bool:
        """
         ASCII码表中的全部的特殊字符
        Args:
            context:
        Returns:
        Examples:
            >>> examples = ["[", ".", "^", "&3%", "1235678", "abc", "ABC", "ABCa123"]
            >>> BatchTask.list_jobs(RegEx.match_ascii_special_char, examples)
        """
        return bool(re.match("[!-/:-@[-`{-~]+", str(context)))
